"""A collection of stand-alone functions to extract information from the CasADi
documentation via code. In particular, this module offers a way to get the solvers that
are available in CasADi (i.e., they have an interface) as well as and their options. The
functions are taken from the
`MPCTools <https://bitbucket.org/rawlings-group/mpc-tools-casadi/src/master/mpctools/util.py>`_
repository by the Rawlings' group.
"""

import contextlib
import itertools
import warnings
from typing import Any as _Any
from typing import Callable
from typing import NamedTuple as _NamedTuple

import casadi as cs


class _LambdaType:
    def __init__(self, func: Callable[[_Any], _Any], typerepr: str) -> None:
        self.__typerepr = typerepr
        self.__func = func

    def __call__(self, val: _Any) -> _Any:
        return self.__func(val)

    def __repr__(self) -> str:
        return f"<type '{self.__typerepr}'>"

    def __str__(self) -> str:
        return repr(self)


class _DocCell(_NamedTuple):
    id: str
    default: str
    doc: str


_TABLE_START = "+="  # string prefix that starts the table
_CELL_END = "+-"  # string prefix that ends the cell
_CELL_CONTENTS = "|"  # string prefix that continues the cell
_TYPES: dict[str, Callable] = {  # casadi types to python types
    "OT_INTEGER": int,
    "OT_STRING": str,
    "OT_REAL": float,
    "OT_INT": int,
    "OT_DICT": dict,
    "OT_DOUBLE": float,
    "OT_BOOL": bool,
    "OT_STR": str,
    "OT_INTVECTOR": _LambdaType(lambda x: [int(i) for i in x], "list[int]"),
    "OT_STRINGVECTOR": _LambdaType(lambda x: [str(i) for i in x], "list[str]"),
}


def _get_doc_cell(lines: list[str]) -> _DocCell:
    """Returns a DocCell tuple for the set of lines.

    joins is a tuple of strings to say how to join multiple lines in a given
    cell. It must have exactly one entry for each cell.
    """
    ncol = lines[0].count(" | ") + 1
    assert ncol in (3, 4), f"Expected 3 or 4 columns in the docstring table; got {ncol}"

    fields: tuple[list[str], ...] = tuple([] for _ in range(ncol))
    for line in lines:
        cells = [c.strip() for c in line.split(" | ") if c]
        cells[0] = cells[0].lstrip("|").rstrip()
        cells[-1] = cells[-1].rstrip("|").rstrip()
        for i, c in enumerate(cells):
            fields[i].append(c.strip())
            if i == ncol - 1:
                break
        else:
            raise ValueError("Wrong number of columns.")

    if ncol == 3:
        id, type, doc = (j.join(f) for j, f in zip(("", "", " "), fields))
        default = None

    else:
        id, type, default, doc = (j.join(f) for j, f in zip(("", "", "", " "), fields))
        if typefunc := _TYPES.get(type):
            try:
                default = typefunc(default)
            except (ValueError, TypeError):
                includetype = True
                if default in ("None", "GenericType()"):
                    default = None
                elif typefunc is int:
                    with contextlib.suppress(ValueError, TypeError):
                        default = int(float(default))
                        includetype = False
                if includetype:
                    default = (default, typefunc)
        else:
            warnings.warn(f"Unknown type for '{id}', '{type}'.", stacklevel=2)

    return _DocCell(id, default, doc)


def _get_doc_dict(docstring: str) -> dict[str, tuple[_Any, str]]:
    lineiter = itertools.dropwhile(
        lambda x: not x.startswith(_TABLE_START), docstring.split("\n")
    )
    try:
        next(lineiter)
    except StopIteration as e:
        raise ValueError("No table found!") from e
    thiscell: list[str] = []
    allcells: list[_DocCell] = []
    for line in lineiter:
        if line.startswith(_CELL_END):
            allcells.append(_get_doc_cell(thiscell))
            thiscell.clear()
        elif line.startswith(_CELL_CONTENTS):
            thiscell.append(line)
        else:
            break
    return {c.id: ("N/A" if c.default is None else c.default, c.doc) for c in allcells}


def get_casadi_plugins() -> dict[str, list[str]]:
    """Returns the available CasADi plugins.

    Returns
    -------
    dict of (str, list of str)
        A dictionary containing for each type of problem a list of plugin names that
        are available.

    Raises
    ------
    RuntimeError
        Raises in case the plugins cannot be retrieved because the functions
        :func:`cs.CasadiMeta_getPlugins` or :func:`cs.CasadiMeta_plugins` are not
        available.
    """

    func = getattr(cs, "CasadiMeta_getPlugins", getattr(cs, "CasadiMeta_plugins", None))
    if func is None:
        raise RuntimeError("Unable to get Casadi plugins.")

    all_plugins: str = func()
    plugins = (p.split("::") for p in all_plugins.split(";"))
    plugin_dict: dict[str, list[str]] = {}
    for group, name in plugins:
        if group not in plugin_dict:
            plugin_dict[group] = []
        plugin_dict[group].append(name)
    return plugin_dict


def list_available_solvers() -> dict[str, list[str]]:
    """Returns the available CasADi solvers.

    Returns
    -------
    dict of (str, list of str)
        A dictionary containing for each type of problem a list of plugin names that
        are available.

    Raises
    ------
    RuntimeError
        Raises in case the plugins cannot be retrieved because the functions
        :func:`cs.CasadiMeta_getPlugins` or :func:`cs.CasadiMeta_plugins` are not
        available.
    """
    availablesolvers = get_casadi_plugins()
    return {
        "nlp": availablesolvers.get("Nlpsol", []),
        "qp": availablesolvers.get("Conic", []) + availablesolvers.get("Qpsol", []),
    }


def get_solver_options(
    solver: str, display: bool = True
) -> dict[str, tuple[_Any, str]]:
    """Returns the solver-specific options, with default value and description whenever
    available.

    Parameters
    ----------
    solver : str
        The solver name.
    display : bool, optional
        Whether to print the options, by default ``True``

    Returns
    -------
    dict of (str, tuple of (Any, str)
        A dictionary containing for each option the default value and a description.

    Raises
    ------
    ValueError
        Raises in case ``solver`` is not included in the available solvers.
    """
    availablesolvers = list_available_solvers()
    if solver in availablesolvers["nlp"]:
        docstring = cs.doc_nlpsol(solver)
    elif solver in availablesolvers["qp"]:
        docstring = cs.doc_conic(solver)
    else:
        raise ValueError(f"Unknown solver: '{solver}'.")
    options = _get_doc_dict(docstring)
    if display:
        print(f"Available options [default] for {solver}:\n")
        for k in sorted(options.keys()):
            default, doc = options[k]
            print(f"{k} [{default!r}]: {doc}\n")
    return options



from collections.abc import Collection, Iterable, Sequence
from dataclasses import dataclass
from numbers import Integral
from typing import Any, Literal, Optional, TypeVar, Union

import casadi as cs
import numpy as np
import numpy.typing as npt

from csnlp.core.solutions import Solution

from ...core.data import find_index_in_vector
from .mpc import Mpc

SymType = TypeVar("SymType", cs.SX, cs.MX)


def _n(parname: str, prefix: str) -> str:
    """Internal utility for the naming convention of the TVA dynamics's parameters."""
    return f"{parname}__{prefix}"


def _create_atv_mats(
    N: int, As: list[SymType], Bs: list[SymType], C: SymType
) -> tuple[SymType, SymType, SymType]:
    """Internal utility to build the affine time-varying (ATV) matrices."""
    A0 = As[0]
    ns = A0.size1()

    row = eye = cs.DM.eye(ns)
    rows = [cs.horzcat(row, cs.DM(ns, ns * (N - 1)))]
    for k, A in enumerate(As[1:], start=1):
        row = cs.horzcat(A @ row, eye)
        rows.append(cs.horzcat(row, cs.DM(ns, ns * (N - k - 1))))
    base = cs.vcat(rows)

    F = base[:, :ns] @ A0
    G = base @ cs.dcat(Bs)
    L = base @ C
    return F, G, L


@dataclass
class PwaRegion:
    """Stores the matrices defining the i-th region in a piecewise affine system."""

    A: npt.NDArray[np.floating]
    r"""The state matrix :math:`A_i` of the affine dynamics
    :math:`x_+ = A_i x + B_i u + c_i`."""

    B: npt.NDArray[np.floating]
    r"""The input matrix :math:`B_i` of the affine dynamics
    :math:`x_+ = A_i x + B_i u + c_i`."""

    c: npt.NDArray[np.floating]
    r"""The affine constant vector :math:`c_i` of the affine dynamics
    :math:`x_+ = A_i x + B_i u + c_i`."""

    S: npt.NDArray[np.floating]
    r"""The state matrix :math:`S_i` of the region inequality
    :math:`S_i [x^\top, u^\top]^\top \leq T_i`"""

    T: npt.NDArray[np.floating]
    r"""The constant vector :math:`T_i` of the region inequality
    :math:`S_i x + R_i u \leq T_i`"""


class PwaMpc(Mpc[SymType]):
    r"""MPC controller for piecewise affine (PWA) systems. A PWA system is characterized
    by linear dynamics that switch between different regions of the state-action space.
    In mathematical terms, given a PWA system with :math:`s` regions, the dynamics are

    .. math::
        x_+ = \begin{cases}
            A_1 x + B_1 u + c_1 & \text{if } S_1 x + R_1 u \leq T_1 \\
            & \vdots \\
            A_i x + B_i u + c_i & \text{if } S_i x + R_i u \leq T_i \\
            & \vdots \\
            A_s x + B_s u + c_s & \text{if } S_s x + R_s u \leq T_s
        \end{cases}

    This MPC controller class can then handle two different types of specifications for
    the dynamics.

    * **affine time-varying dynamics**: the MPC controller can be considered as a
      linear MPC controller with time-varying dynamics, in which case the dynamics
      must be defined via the :meth:`set_affine_time_varying_dynamics` method. Then,
      prior to solving the optimization problem, the sequence of regions to be active
      at each time-step needs to be set via :meth:`set_switching_sequence` by the
      user/externally.
    * **optimzing also the sequence**: alternatively, the sequence of PWA regions can
      be optimized over, in which case the dynamics must be defined via the
      :meth:`set_pwa_dynamics` method. Following :cite:`bemporad_control_1999`, the PWA
      dynamics are converted to mixed-logical dynamical form, and the ensuing MPC
      optimization becomes a mixed-integer optimization problem. This is done under the
      hood via the :meth:`set_pwa_dynamics` method. See also
      :cite:`borrelli_predictive_2017` for further details.

    Parameters
    ----------
    nlp : Nlp
        NLP scheme to be wrapped
    prediction_horizon : int
        A positive integer for the prediction horizon of the MPC controller.
    control_horizon : int, optional
        A positive integer for the control horizon of the MPC controller. If not given,
        it is set equal to the control horizon.
    input_spacing : int, optional
        Spacing between independent input actions. This argument allows to reduce the
        number of free actions along the control horizon by allowing only the first
        action every ``n`` to be free, and the following ``n-1`` to be fixed equal to
        that action (where ``n`` is given by ``input_spacing``). By default, no spacing
        is allowed, i.e., ``1``.
    shooting : 'single' or 'multi', optional
        Type of approach in the direct shooting for parametrizing the control
        trajectory. See Section 8.5 in :cite:`rawlings_model_2017`. By default, direct
        shooting is used.

    Raises
    ------
    ValueError
        Raises if the shooting method is invalid; or if any of the horizons are invalid;
        or if the number of scenarios is not a positive integer."""

    tva_dynamics_name = "tva_dyn"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._fixed_sequence_dynamics = False
        self._pwa_system: Optional[Sequence[PwaRegion]] = None
        self._sequence: Optional[Collection[int]] = None

    def validate_pwa_dimensions(self, pwa_system: Iterable[PwaRegion]) -> None:
        """Validates that the dimensions are correct for all matrices in
        the passed PWA system.

        Parameters
        ----------
        pwa_system : iterable of PwaRegion
            An iterable of :class:`PwaRegion` objects, where the i-th object contains
            the matrices defining the i-th region of the PWA system.

        Raises
        ------
        ValueError
            Raises if the dimensions of any matrix in any region do not match the
            expected shape.
        """
        ns = self.ns
        na = self.na
        nsa = ns + na
        n_ineq = pwa_system[0].T.shape[0]  # must be the same for all regions
        for i, region in enumerate(pwa_system):
            if region.A.shape != (ns, ns):
                raise ValueError(f"A in region {i} must have shape ({ns}, {ns}).")
            if region.B.shape != (ns, na):
                raise ValueError(f"B in region {i} must have shape ({ns}, {na}).")
            if region.c.shape != (ns,):
                raise ValueError(f"c in region {i} must have shape ({ns},).")
            if region.S.shape != (n_ineq, nsa):
                raise ValueError(f"S in region {i} must have shape ({n_ineq}, {nsa}).")
            if region.T.shape != (n_ineq,):
                raise ValueError(f"T in region {i} must have shape ({n_ineq},).")

    def set_pwa_dynamics(
        self,
        pwa_system: Collection[PwaRegion],
        D: Union[npt.NDArray[np.floating], cs.DM],
        E: npt.NDArray[np.floating],
        clp_opts: Optional[dict[str, Any]] = None,
        parallelization: Literal[
            "serial", "unroll", "inline", "thread", "openmp"
        ] = "thread",
        max_num_threads: Optional[int] = None,
    ) -> None:
        r"""Sets the piecewise affine dynamics of the system for the MPC controller,
        creating auxiliary variables and constraints to handle the PWA switching. In
        order to perform the conversion of the PWA dynamics to mixed-logical dynamical
        form, the method solves a series of linear programmes via the ``CLP`` solver.
        Parallelization can also be enabled to speed up the process.

        Parameters
        ----------
        pwa_system : collection of PwaRegion
            A collection of :class:`PwaRegion` objects, where the i-th object contains
            the matrices defining the i-th region of the PWA system.
        D : array or casadi.DM of shape (n_ineq, ns + na)
            The (possibly sparse) matrix ``D`` defining the polytopic constraints on the
            state-action space :math:`D [x^\top, u^\top]^\top \leq E`, where ``ns`` and
            ``na`` are the numbers of states and actions in the MPC problem,
            respectively.
        E : array of shape (n_ineq,)
            The matrix ``E`` defining the polytopic constraints on the state-action
            space :math:`D [x^\top, u^\top]^\top \leq E`.
        clp_opts : dict, optional
            Options for the CLP solver. Defaults to ``None``.
        parallelization : "serial", "unroll", "inline", "thread", "openmp"
            The type of parallelization to use (see :func:`casadi.Function.map`) when
            solving the linear programmes. By default, ``"thread"`` is selected.
        max_num_threads : int, optional
            Maximum number of threads to use in parallelization; if ``None``, the number
            of threads is equal to the number of regions in the system.

        Raises
        ------
        RuntimeError
            Raises if the dynamics were already set, or if lower and upper bounds on the
            states and/or actions are set.
        ValueError
            Raises if the dimensions of any matrix in any region do not match the
            expected shape.

        Notes
        -----
        When multiple states and/or control inputs are defined, these are concatenated
        into the single vectors :math:`x` and :math:`u`, respectively. Also, this
        function will raise an error if lower and upper bounds on any state or action
        are set. This is because these bounds should be instead specified via the
        matrices ``D`` and ``E``. Moreover, this function only uses these matrices for
        internal computations, so the user should take care to impose the constraints
        :math:`D [x^\top, u^\top]^\top \leq E` in the optimization problem, as well as
        any other, via the :meth:`constraint` method.
        """
        if self._dynamics_already_set:
            raise RuntimeError("Dynamics were already set.")

        # retrieve lower and upper bounds on the states and actions and check they are
        # all unset, i.e., -/+ inf
        U = cs.vvcat(self._actions_exp.values())
        idx = find_index_in_vector(self.x, U)
        lbu = self.lbx[idx].data
        ubu = self.ubx[idx].data
        if (~np.isneginf(lbu)).any() or (~np.isposinf(ubu)).any():
            raise RuntimeError(
                "Cannot set lower and upper bounds on the actions in PWA systems; use "
                "arguments `D` and `E` of `set_pwa_dyanmics` instead."
            )
        if self._is_multishooting:
            X = cs.vvcat(self._states.values())
            idx = find_index_in_vector(self.x, X)
            lbx = self.lbx[idx].data
            ubx = self.ubx[idx].data
            if (~np.isneginf(lbx)).any() or (~np.isposinf(ubx)).any():
                raise RuntimeError(
                    "Cannot set lower and upper bounds on the states in PWA systems; "
                    "use arguments `D` and `E` of `set_pwa_dyanmics` instead."
                )

        # validate dimensions
        self.validate_pwa_dimensions(pwa_system)
        nsa = self.ns + self.na
        n_ineq = E.shape[0]
        if D.shape != (n_ineq, nsa):
            raise ValueError(f"D must have shape ({n_ineq}, {nsa}).")
        if E.shape != (n_ineq,):
            raise ValueError(f"E must have shape ({n_ineq},).")

        # set the PWA dynamics
        if max_num_threads is None:
            max_num_threads = len(pwa_system)
        if clp_opts is None:
            clp_opts = {}
        self._set_pwa_dynamics(
            pwa_system, D, E, clp_opts, parallelization, max_num_threads
        )
        self._dynamics_already_set = True

    def set_affine_time_varying_dynamics(
        self, pwa_system: Sequence[PwaRegion]
    ) -> tuple[Optional[SymType], Optional[SymType], Optional[SymType]]:
        r"""Sets affine time-varying dynamics as the controller's prediction model and
        creates the corresponding dynamics constraints. The dynamics in the affine
        time-varying form are described as

        .. math:: x_{k+1} = A_k x_k + B_k u_k + c_k.

        where :math:`x_k` and :math:`x_{k+1}` are the current and next state,
        :math:`u_k` is the control action, :math:`w_k` is the disturbance, and
        :math:`A_k, B_k, c_k` are the constant matrices of the region to be visited by
        the state trajectory at time step `k`. By setting the dynamics with this method,
        the user can then specify the switching sequence of regions to be active at each
        time step along the prediction horizon via :meth:`set_switching_sequence`, and
        solve a much simpler optimization problem, as the sequence is fixed. Instead, if
        :meth:`set_pwa_dynamics` is used, the sequence of regions is optimized over via
        a (usually expensive) mixed-integer optimization problem.

        Parameters
        ----------
        pwa_system : sequence of PwaRegion
            A sequence of :class:`PwaRegion` objects, where the i-th object contains
            the matrices defining the i-th region of the PWA system.

        Returns
        -------
        Optional 3-tuple of symbolic or numerical arrays
            In multiple shooting, returns a tuple of ``None``s. In single shooting,
            returns the matrices :math:`F, G, L` that parametrize the dynamics. See,
            e.g., :cite:`campi_scenario_2019`.

        Raises
        ------
        RuntimeError
            Raises if the dynamics were already set.
        ValueError
            Raises if the dimensions of any matrix in any region do not match the
            expected shape.
        """
        if self._dynamics_already_set:
            raise RuntimeError("Dynamics were already set.")

        self.validate_pwa_dimensions(pwa_system)
        N = self._prediction_horizon
        ns = self.ns
        na = self.na
        nsa = ns + na
        n_ineq = pwa_system[0].T.size

        prefix = self.tva_dynamics_name
        As = cs.vertsplit_n(self.parameter(_n("A", prefix), (ns * N, ns)), N)
        Bs = cs.vertsplit_n(self.parameter(_n("B", prefix), (ns * N, na)), N)
        C = self.parameter(_n("c", prefix), (ns * N, 1))
        Cs = cs.vertsplit_n(C, N)
        Ss = cs.vertsplit_n(self.parameter(_n("S", prefix), (n_ineq * N, nsa)), N)
        T = self.parameter(_n("T", prefix), (n_ineq * N, 1))

        if self._is_multishooting:
            X, U = self._set_multishooting_tva_dynamics(As, Bs, Cs)
            F = G = L = None
        else:
            X, U, F, G, L = self._set_singleshooting_tva_dynamics(As, Bs, C)

        S = cs.dcat(Ss)
        XU = cs.vec(cs.vertcat(X[:, :-1], U))  # NOTE: different from vvcat!
        self.constraint("region", S @ XU - T, "<=", 0)

        self._pwa_system = pwa_system
        self._dynamics_already_set = True
        self._fixed_sequence_dynamics = True
        return F, G, L

    def set_switching_sequence(self, sequence: Collection[int]) -> None:
        """Sets the sequence of regions to be active at each time step along the MPC
        prediction horizon. Then, when solved, the MPC optimization will only optimze
        over the action and state trajectories, while enforcing the sequence of regions
        visited by the states to be the one provided here.

        Parameters
        ----------
        sequence : collection of int
            A collection of integers representing the indices of the regions to be
            active at each time step along the prediction horizon, i.e., the k-th entry
            of this collection represents the index of the region the state must be at
            time step ``k``.

        Raises
        ------
        ValueError
            Raises if dynamics have not been set via
            :meth:`set_affine_time_varying_dynamics`; if the sequence is not the same
            length as the prediction horizon; if the sequence does not contain integers;
            if the sequence contains integers that exceed the number of PWA regions
            specified via :meth:`set_affine_time_varying_dynamics`.

        Notes
        -----
        For internal validation purposes, please call first
        :meth:`set_affine_time_varying_dynamics` and only then call this method.
        """
        if not self._fixed_sequence_dynamics or self._pwa_system is None:
            raise ValueError(
                "Sequence can only be set if time-varying dynamics are used."
            )
        if len(sequence) != self._prediction_horizon:
            raise ValueError("Length of sequence must match the prediction horizon")

        N = len(self._pwa_system)
        for i, idx in enumerate(sequence):
            if not isinstance(idx, Integral):
                raise ValueError(
                    f"{i}-th element of the sequence is not an integer; got {idx}."
                )
            if not 0 <= idx < N:
                raise ValueError(
                    f"{i}-th element of the sequence must be in [0, N); got {idx}."
                )
        self._sequence = sequence

    def solve(
        self,
        pars: Optional[dict[str, npt.ArrayLike]] = None,
        vals0: Optional[dict[str, npt.ArrayLike]] = None,
    ) -> Solution[SymType]:
        if self._fixed_sequence_dynamics:
            regions = self._pwa_system
            assert regions is not None, "PWA system should have been set!"
            if self._sequence is None:
                raise ValueError(
                    "A sequence must be set via `set_switching_sequence` prior to "
                    "solving the MPC because the dyanmics were set via "
                    "`set_affine_time_varying_dynamics`. Use `set_pwa_dynamics` instead"
                    " to optimize over the sequence as well."
                )

            if pars is None:
                pars = {}
            As = []
            Bs = []
            Cs = []
            Ss = []
            Ts = []
            for idx in self._sequence:
                As.append(regions[idx].A)
                Bs.append(regions[idx].B)
                Cs.append(regions[idx].c)
                Ss.append(regions[idx].S)
                Ts.append(regions[idx].T)
            prefix = self.tva_dynamics_name
            pars[_n("A", prefix)] = np.concatenate(As, 0)
            pars[_n("B", prefix)] = np.concatenate(Bs, 0)
            pars[_n("c", prefix)] = np.concatenate(Cs, 0)
            pars[_n("S", prefix)] = np.concatenate(Ss, 0)
            pars[_n("T", prefix)] = np.concatenate(Ts, 0)
        return self.nlp.solve(pars, vals0)

    @staticmethod
    def get_optimal_switching_sequence(
        sol: Solution[SymType],
    ) -> npt.NDArray[np.integer]:
        """Returns the optimal switching sequence of regions for the state trajectory
        along the prediction horizon, which can be extracted from the solution of the
        corresponding mixed-integer optimization MPC problem (i.e., when the dynamics
        are set via :meth:`set_pwa_dynamics`, which allows to optimize over the sequence
        as well, though more computationally expensive).

        Parameters
        ----------
        sol : Solution
            An optimal solution of the mixed-integer PWA MPC problem.

        Returns
        -------
        array of ints
            An array of integers representing the indices of the regions that active at
            each time step along the prediction horizon, i.e., the k-th entry of this
            array represents the index of the region the predicted state is in at time
            step ``k``.

        Raises
        ------
        KeyError
            Raises if the solution does not contain the variable ``delta``, e.g., the
            solution was not obtained from a mixed-integer optimization problem, or the
            dynamics were not set via :meth:`set_pwa_dynamics`.
        """
        return sol.vals["delta"].toarray().argmax(0)

    def _set_pwa_dynamics(
        self,
        regions: Collection[PwaRegion],
        D: Union[npt.NDArray[np.floating], cs.DM],
        E: npt.NDArray[np.floating],
        clp_opts: dict[str, Any],
        parallelization: Literal["serial", "unroll", "inline", "thread", "openmp"],
        max_num_threads: int,
    ) -> None:
        """Internal utility to create PWA dynamics constraints."""
        # solve linear programs to determine bounds for big-M relaxations. These LPs are
        # solved parallelly for each region and for each inequality defining the region.
        D = cs.sparsify(D)  # can be sparse
        S_, T_, AB_, C_ = [], [], [], []
        for r in regions:
            S_.append(r.S)
            T_.append(r.T)
            AB_.append(np.hstack((r.A, r.B)))
            C_.append(r.c)
        S = np.vstack(S_).T
        T = np.asarray(T_)
        AB = np.vstack(AB_).T
        C = np.asarray(C_)

        nr = len(regions)
        n_ineq = r.T.size
        ns, na = r.B.shape
        nsa = ns + na
        N = self._prediction_horizon

        lp = {"h": cs.Sparsity(nsa, nsa), "a": D.sparsity()}
        lpsolver = cs.conic("lpsolver", "clp", lp, clp_opts)

        mapped_lpsolver = lpsolver.map(nr * n_ineq, parallelization, max_num_threads)
        sol = mapped_lpsolver(g=S, a=D, uba=E)
        big_M = -sol["cost"].toarray().reshape(nr, n_ineq) + T

        mapped_lpsolver = lpsolver.map(nr * ns, parallelization, max_num_threads)
        sol = mapped_lpsolver(g=AB, a=D, uba=E)
        tmp_lb = sol["cost"].toarray().reshape(nr, ns) + C
        M_lb = tmp_lb.min(0)
        sol = mapped_lpsolver(g=-AB, a=D, uba=E)
        tmp_ub = -sol["cost"].toarray().reshape(nr, ns) + C
        M_ub = tmp_ub.max(0)

        # dynamics constraints - we now have to add constraints for all regions at each
        # time-step, with the binary variable delta selecting the active region
        z = [self.variable(f"z_{i}", (ns, N))[0] for i in range(nr)]
        if self._is_multishooting:
            X = cs.vcat(self._states.values())
            self.constraint("dynamics", X[:, 1:], "==", sum(z))
        else:
            Xk = cs.vcat(self._initial_states.values())
            X = cs.horzcat(Xk, sum(z))
            cumsizes = np.cumsum(
                [0] + [s.shape[0] for s in self._initial_states.values()]
            )
            self._states = dict(zip(self._states.keys(), cs.vertsplit(X, cumsizes)))

        U = cs.vcat(self._actions_exp.values())
        delta, _, _ = self.variable("delta", (nr, N), lb=0, ub=1, discrete=True)
        self.constraint("delta_sum", cs.sum1(delta), "==", 1)
        z_ub = []
        z_lb = []
        region = []
        z_x_ub = []
        z_x_lb = []
        XU = cs.vertcat(X[:, :-1], U)
        for i, r in enumerate(regions):
            z_i = z[i]
            delta_i = delta[i, :]
            AB_i = AB_[i]
            z_ub.append(z_i - M_ub @ delta_i)
            z_lb.append(z_i - M_lb @ delta_i)
            region.append(r.S @ XU - r.T - big_M[i, :] @ (1 - delta_i))
            z_x_ub.append(z_i - (AB_i @ XU + r.c - M_lb @ (1 - delta_i)))
            z_x_lb.append(z_i - (AB_i @ XU + r.c - M_ub @ (1 - delta_i)))
        self.constraint("z_ub", cs.vcat(z_ub), "<=", 0)
        self.constraint("z_lb", cs.vcat(z_lb), ">=", 0)
        self.constraint("region", cs.vcat(region), "<=", 0)
        self.constraint("z_x_ub", cs.vcat(z_x_ub), "<=", 0)
        self.constraint("z_x_lb", cs.vcat(z_x_lb), ">=", 0)

    def _set_multishooting_tva_dynamics(
        self, As: Iterable[SymType], Bs: Iterable[SymType], Cs: Iterable[SymType]
    ) -> tuple[SymType, SymType]:
        """Internal utility to create affine time-varying dynamics constraints in
        multiple shooting."""
        X = cs.vcat(self._states.values())
        U = cs.vcat(self._actions_exp.values())
        X_next_ = []
        for k, (A, B, c) in enumerate(zip(As, Bs, Cs)):
            X_next_.append(A @ X[:, k] + B @ U[:, k] + c)
        X_next = cs.hcat(X_next_)
        self.constraint("dynamics", X_next, "==", X[:, 1:])
        return X, U

    def _set_singleshooting_tva_dynamics(
        self, As: Iterable[SymType], Bs: Iterable[SymType], C: SymType
    ) -> tuple[SymType, SymType, SymType, SymType, SymType]:
        """Internal utility to create affine time-varying dynamics constraints in
        mutple shooting."""
        N = self._prediction_horizon
        F, G, L = _create_atv_mats(self._prediction_horizon, As, Bs, C)
        x_0 = cs.vcat(self._initial_states.values())
        U = cs.vcat(self._actions_exp.values())
        X_next = F @ x_0 + G @ cs.vec(U) + L

        # append initial state, reshape and save to internal dict
        ns = F.size2()
        X = cs.vertcat(x_0, X_next).reshape((ns, N + 1))
        cumsizes = np.cumsum([0] + [s.shape[0] for s in self._initial_states.values()])
        self._states = dict(zip(self._states.keys(), cs.vertsplit(X, cumsizes)))
        return X, U, F, G, L



from math import ceil
from typing import Callable, Literal, Optional, TypeVar, Union

import casadi as cs
import numpy as npy
import numpy.typing as npt

from csnlp.multistart.multistart_nlp import _chained_substitute, _n

from ...util.math import repeat
from ..wrapper import Nlp
from .mpc import Mpc, _create_ati_mats
from .scenario_based_mpc import ScenarioBasedMpc

SymType = TypeVar("SymType", cs.SX, cs.MX)
MatType = TypeVar("MatType", SymType, cs.DM, npy.ndarray)


class MultiScenarioMpc(ScenarioBasedMpc[SymType]):
    """Multi-scenario model predictive control (MSMPC) wrapper for a :class:`csnlp.Nlp`
    scheme. It generlizes :class:`csnlp.wrappers.ScenarioBasedMpc` to allow for multiple
    scenarios, each with its own state trajectory, action sequence (a number of these
    actions can be shared across all scenarios; see ``input_sharing``), and parameters.

    Parameters
    ----------
    nlp : Nlp
        NLP scheme to be wrapped
    n_scenarios : int
        Number of scenarios to be considered in the scenario-based MPC formulation. Must
        be a positive integer.
    prediction_horizon : int
        A positive integer for the prediction horizon of the MPC controller.
    control_horizon : int, optional
        A positive integer for the control horizon of the MPC controller. If not given,
        it is set equal to the control horizon.
    input_spacing : int, optional
        Spacing between independent input actions. This argument allows to reduce the
        number of free actions along the control horizon by allowing only the first
        action every ``n`` to be free, and the following ``n-1`` to be fixed equal to
        that action (where ``n`` is given by ``input_spacing``). By default, no spacing
        is allowed, i.e., ``1``.
    input_sharing : int, optional
        Number of actions that are shared across all scenarios. If ``0``, each scenario
        has its own action trajectory; if ``n``, all scenarios share the same first
        ``n`` actions. Must not be larger than the number of free actions in the
        control horizon. By default, ``1``, so only the first action is shared across
        the scenarios.
    shooting : 'single' or 'multi', optional
        Type of approach in the direct shooting for parametrizing the control
        trajectory. See Section 8.5 in :cite:`rawlings_model_2017`. By default, direct
        shooting is used.

    Raises
    ------
    ValueError
        Raises if the shooting method is invalid; or if any of the horizons are invalid;
        or if the number of scenarios is not a positive integer; or if
        ``input_sharing`` is not a valid integer (nonnegative and not larger than the
        number of free actions in the control horizon).
    """

    def __init__(
        self,
        nlp: Nlp[SymType],
        n_scenarios: int,
        prediction_horizon: int,
        control_horizon: Optional[int] = None,
        input_spacing: int = 1,
        input_sharing: int = 1,
        shooting: Literal["single", "multi"] = "multi",
    ) -> None:
        super().__init__(
            nlp,
            n_scenarios,
            prediction_horizon,
            control_horizon,
            input_spacing,
            shooting,
        )

        nu_free = ceil(self.control_horizon / self._input_spacing)
        if not isinstance(input_sharing, (int, npy.integer, npy.int_)) or not (
            0 <= input_sharing <= nu_free
        ):
            raise ValueError(
                f"`input_sharing` in range [0, {nu_free}]; got {input_sharing} instead."
            )
        self._input_sharing = input_sharing

        self.single_parameters: dict[str, SymType] = {}
        self.single_actions: dict[str, SymType] = {}
        self.single_actions_exp: dict[str, SymType] = {}

    @property
    def na(self) -> int:
        """Number of actions in a single scenario in the NLP scheme."""
        return super().na // self._n_scenarios

    @property
    def na_all(self) -> int:
        """Gets the number of actions of the MSMPC controller considering all
        scenarios."""
        return super().na

    @property
    def np(self) -> int:
        """Number of parameters in a single scenario in the NLP scheme."""
        return self.nlp.np // self._n_scenarios

    @property
    def np_all(self) -> int:
        """Gets the number of parameters of the MSMPC controller considering all
        scenarios."""
        return self.nlp.np

    def parameters_i(self, i: int) -> dict[str, SymType]:
        """Gets the symbolic parameters belonging to the i-th scenario."""
        return {n: self.parameters[_n(n, i)] for n in self.single_parameters}

    def actions_i(self, i: int) -> dict[str, SymType]:
        """Gets the symbolic actions belonging to the i-th scenario."""
        return {n: self.actions[_n(n, i)] for n in self.single_actions_exp}

    def actions_expanded_i(self, i: int) -> dict[str, SymType]:
        """Gets the symbolic expanded actions belonging to the i-th scenario."""
        return {n: self.actions_expanded[_n(n, i)] for n in self.single_actions_exp}

    def parameter(
        self, name: str, shape: tuple[int, int] = (1, 1)
    ) -> tuple[SymType, list[SymType]]:
        """Adds one parameter per scenario to the MSMPC controller.

        Parameters
        ----------
        name : str
            Name of the new parameter. Must not be already in use.
        shape : tuple of 2 ints, optional
            Shape of the new parameter. By default a scalar, i.e., ``(1, 1)``.

        Returns
        -------
        single_parameter : casadi.SX or MX
            Symbol corresponding to the parameter of a single scenario. See the note
            for :meth:``state``
        parameters : list of casadi.SX or MX
            The symbols for the new parameters of each scenario in the MSMPC controller.

        Raises
        ------
        ValueError
            Raises if there is already another parameter with the same name ``name``.
        """
        ps = []
        for i in range(self._n_scenarios):
            name_i = _n(name, i)
            p_i = self.nlp.parameter(name_i, shape)
            ps.append(p_i)
        p_single = self.nlp._sym_type.sym(name, shape)
        self.single_parameters[name] = p_single
        return p_single, ps

    def action(
        self,
        name: str,
        size: int = 1,
        discrete: bool = False,
        lb: Union[npt.ArrayLike, cs.DM] = -npy.inf,
        ub: Union[npt.ArrayLike, cs.DM] = +npy.inf,
    ) -> tuple[SymType, SymType, list[SymType], list[SymType]]:
        """Adds one control action variable per scenario to the MSMPC controller.
        Automatically handles sharing of actions across scenarios, generating less free
        actions if some of these are to be shared across scenarios.

        Parameters
        ----------
        name : str
            Name of the control action.
        size : int, optional
            Size of the control action (assumed to be a vector). Defaults to ``1``.
        discrete : bool, optional
            Flag indicating if the action is discrete. Defaults to ``False``.
        lb : array_like, casadi.DM, optional
            Hard lower bound of the control action, by default ``-np.inf``.
        ub : array_like, casadi.DM, optional
            Hard upper bound of the control action, by default ``+np.inf``.

        Returns
        -------
        single action : casadi.SX or MX
            Symbol corresponding to the action of a single scenario. This is useful for
            automatically defining, e.g., the objective and constraints over the various
            scenarios of the MSMPC controller, but it is not used in the actual NLP
            solver.
        single action expanded : casadi.SX or MX
            Same as above, but the action is expanded to the same length of the
            prediction horizon.
        actions : list of casadi.SX or MX
            The list of the action symbolic variable.
        actions expanded : list of casadi.SX or MX
            The list of the same action symbolic variable, but expanded to the same
            length of the prediction horizon.
        """
        Np = self._prediction_horizon
        Nc = self._control_horizon
        spacing = self._input_spacing
        nu_free = ceil(Nc / spacing)
        sharing = self._input_sharing
        shape = (size, nu_free)

        def _expand(u: SymType) -> SymType:
            u_exp = u if spacing == 1 else repeat(u, (1, spacing))[:, :Nc]
            if gap := Np - u_exp.shape[1]:
                last = u_exp[:, -1]
                u_exp = cs.horzcat(u_exp, *(last for _ in range(gap)))
            return u_exp

        us = []
        us_exp = []
        if sharing == 0:
            # no sharing, each scenario has its own action trajectory
            for i in range(self._n_scenarios):
                name_i = _n(name, i)
                u_i = self.nlp.variable(name_i, shape, discrete, lb, ub)[0]
                u_exp_i = _expand(u_i)
                us.append(u_i)
                us_exp.append(u_exp_i)
                self._actions[name_i] = u_i
                self._actions_exp[name_i] = u_exp_i

        elif sharing == nu_free:
            # full sharing, all scenarios share the same action trajectory
            u_unique = self.nlp.variable(_n(name, "shared"), shape, discrete, lb, ub)[0]
            u_exp_unique = _expand(u_unique)
            for i in range(self._n_scenarios):
                name_i = _n(name, i)
                us.append(u_unique)
                us_exp.append(u_exp_unique)
                self._actions[name_i] = u_unique
                self._actions_exp[name_i] = u_exp_unique

        else:
            # partial sharing, each scenario has its own action trajectory, but
            # the first `sharing` actions are shared
            u_shared = self.nlp.variable(
                _n(name, "shared"), (size, sharing), discrete, lb, ub
            )[0]
            shape_not_shared = (size, nu_free - sharing)
            for i in range(self._n_scenarios):
                name_i = _n(name, i)
                u_i = self.nlp.variable(name_i, shape_not_shared, discrete, lb, ub)[0]
                u_i = cs.horzcat(u_shared, u_i)
                u_exp_i = _expand(u_i)
                us.append(u_i)
                us_exp.append(u_exp_i)
                self._actions[name_i] = u_i
                self._actions_exp[name_i] = u_exp_i

        # create also a single symbol for the action and its expanded version
        u_single = self.nlp._sym_type.sym(name, shape)
        u_exp_single = _expand(u_single)
        self.single_actions[name] = u_single
        self.single_actions_exp[name] = u_exp_single
        return u_single, u_exp_single, us, us_exp

    def set_affine_dynamics(
        self,
        A: MatType,
        B: MatType,
        D: Optional[MatType] = None,
        c: Optional[MatType] = None,
        parallelization: Literal[
            "serial", "unroll", "inline", "thread", "openmp"
        ] = "thread",
        max_num_threads: Optional[int] = None,
    ) -> tuple[
        Optional[MatType], Optional[MatType], Optional[MatType], Optional[MatType]
    ]:
        # NOTE: contrary to `ScenarioBasedMpc`, `D` can be optional here. Also, no need
        # to take the number of scenarios into account for threading, as we will run the
        # dynamics once and then chain-substitute for each scenario.
        return Mpc.set_affine_dynamics(
            self, A, B, D, c, parallelization, max_num_threads
        )

    def set_nonlinear_dynamics(
        self,
        F: Union[
            cs.Function,
            Callable[[tuple[npt.ArrayLike, ...]], tuple[npt.ArrayLike, ...]],
        ],
        parallelization: Literal[
            "serial", "unroll", "inline", "thread", "openmp"
        ] = "thread",
        max_num_threads_or_unrolling_base: Optional[int] = None,
    ) -> None:
        # NOTE: contrary to `ScenarioBasedMpc`, `F` is allowed to have no disturbances.
        # Also, no need to take the number of scenarios into account for threading, as
        # we will run the dynamics once and then chain-substitute for each scenario.
        return Mpc.set_nonlinear_dynamics(
            self, F, parallelization, max_num_threads_or_unrolling_base
        )

    def _set_singleshooting_affine_dynamics(
        self, A: MatType, B: MatType, D: Optional[MatType], c: Optional[MatType]
    ) -> tuple[MatType, MatType, Optional[MatType], Optional[MatType]]:
        # NOTE: conversely to `ScenarioBasedMpc`, here first we create the dynamics with
        # the single states/actions/disturbances (we suppose single parameters have been
        # also used in A,B,D,c), and then chain-substitute to get each scenario's
        # dynamics
        ns = A.shape[0]
        N = self._prediction_horizon
        F, G, H, L = _create_ati_mats(self._prediction_horizon, A, B, D, c)
        x_0 = cs.vcat(self._initial_states.values())
        U = cs.vec(
            cs.vcat(self.single_actions_exp.values())  # NOTE: different from vvcat!
        )
        X_next = F @ x_0 + G @ U
        if H is not None:
            X_next += H @ cs.vec(cs.vcat(self.single_disturbances.values()))
        if L is not None:
            X_next += L
        X = cs.vertcat(x_0, X_next).reshape((ns, N + 1))

        state_names = self.single_states.keys()
        cumsizes = npy.cumsum([0] + [s.shape[0] for s in self._initial_states.values()])
        for i in range(self._n_scenarios):
            X_i = self._chained_substitution_for_scenario_i(X, i, True)
            X_i_split = cs.vertsplit(X_i, cumsizes)
            for n, x in zip(state_names, X_i_split):
                self._states[_n(n, i)] = x
        return F, G, H, L

    def _set_multishooting_nonlinear_dynamics(
        self,
        F: cs.Function,
        n_in: int,
        parallelization: Literal["serial", "unroll", "inline", "thread", "openmp"],
        max_num_threads: int,
    ) -> None:
        # NOTE: see note in `_set_singleshooting_affine_dynamics`
        X = cs.vcat(self.single_states.values())
        U = cs.vcat(self.single_actions_exp.values())
        if n_in < 3:
            args = (X[:, :-1], U)
        else:
            D = cs.vcat(self.single_disturbances.values())
            args = (X[:, :-1], U, D)
        Fmap = F.map(self._prediction_horizon, parallelization, max_num_threads)
        X_next = Fmap(*args)

        constraints = []
        for i in range(self._n_scenarios):
            X_i = _chained_substitute(X[:, 1:], (self.single_states, self.states_i(i)))
            X_next_i = self._chained_substitution_for_scenario_i(X_next, i)
            constraints.append(X_i - X_next_i)
        self.constraint("dyn", cs.vvcat(constraints), "==", 0.0)

    def _set_singleshooting_nonlinear_dynamics(
        self, F: cs.Function, n_in: int, base: int
    ) -> None:
        # NOTE: see note in `_set_singleshooting_affine_dynamics`
        X0 = cs.vcat(self._initial_states.values())
        U = cs.vcat(self.single_actions_exp.values())
        if n_in < 3:
            args = (X0, U)
        else:
            D = cs.vcat(self.single_disturbances.values())
            args = (X0, U, D)

        Fmapaccum = F.mapaccum(
            self._prediction_horizon, {"base": base, "allow_free": True}
        )
        X_next = Fmapaccum(*args)
        X = cs.horzcat(X0, X_next)

        state_names = self.single_states.keys()
        cumsizes = npy.cumsum([0] + [s.shape[0] for s in self._initial_states.values()])
        for i in range(self._n_scenarios):
            X_i = self._chained_substitution_for_scenario_i(X, i, True)
            X_i_split = cs.vertsplit(X_i, cumsizes)
            for n, x in zip(state_names, X_i_split):
                self._states[_n(n, i)] = x

    def _chained_substitution_for_scenario_i(
        self, expr: SymType, i: int, skip_states: bool = False
    ) -> Union[SymType, cs.DM]:
        """Iternal utility to perform substitutions in chain for the i-th scenario."""
        states = ({}, {}) if skip_states else (self.single_states, self.states_i(i))
        return _chained_substitute(
            expr,
            (self.single_parameters, self.parameters_i(i)),
            states,
            (self.single_actions, self.actions_i(i)),
            (self.single_disturbances, self.disturbances_i(i)),
            (self.single_slacks, self.slacks_i(i)),
        )



r"""This module contains the core components, aside :class:`csnlp.Nlp` and its base
classes, that are used to build the package.

Overview
========

It contains the following submodules:

- :mod:`csnlp.core.cache`: a collection of methods to handle caching in the package. In
  particular, it offers a decorator :func:`invalidate_cache` that allows to invalidate
  the cache of a given set of other cached properties or methods when the decorated
  method is  invoked, as well as a function :func:`invalidate_caches_of` that allows to
  invalidate the cache of a given object on the fly.
- :mod:`csnlp.core.data`: a collection of functions for manipulating data in CasADi, in
  particular, on how to convert to and from numpy arrays and CasADi symbolic variables,
  and how to find the indices of a symbolic variable in a vector of symbols.
- :mod:`csnlp.core.debug`: scontains classes for storing debug information on the
  parameters, variables and constraints in an instance of the :class:`csnlp.Nlp` class.
- :mod:`csnlp.core.derivatives`: a collection of two methods for computing higher-order
  sensitivities (i.e., Jacobian and Hessian) w.r.t. CasADi symbolic variables. Natively,
  CasADi does not support jacobian or hessian for matrices (or at least, they will be
  flattened). These  "higher-order" functions allows to compute the jacobian and hessian
  of a matrix w.r.t. another matrix.
- :mod:`csnlp.core.scaling`: a collection of classes to perform scaling of variables in
  an :class:`csnlp.Nlp` instance wrapped with :class:`csnlp.wrappers.NlpScaling`. The
  classes in this module inform the wrapper on which variables or parameters to scale
  and how to scale them.
- :mod:`csnlp.core.solutions`: contains classes and methods to store the solution of an
  NLP problem after a call to :meth:`csnlp.Nlp.solve` or
  :meth:`csnlp.multistart.MultistartNlp.solve_multi`.

Submodules
==========

.. autosummary::
   :toctree: generated
   :template: module.rst

   cache
   data
   debug
   derivatives
   scaling
   solutions
"""



from collections.abc import Iterable, Iterator
from functools import lru_cache
from itertools import repeat
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
)

import casadi as cs
import numpy as np
import numpy.typing as npt
from joblib import Parallel, delayed

if TYPE_CHECKING:
    from joblib.memory import MemorizedFunc

from ..core.cache import invalidate_cache
from ..core.solutions import (
    EagerSolution,
    LazySolution,
    Solution,
    _is_infeas,
)
from ..nlps.nlp import Nlp
from ..nlps.objective import _solve_and_get_stats

SymType = TypeVar("SymType", cs.SX, cs.MX)


def _n(sym_name: str, scenario: int) -> str:
    """Internal utility for the naming convention of the ``i``-th scenario's symbols."""
    return f"{sym_name}__{scenario}"


def _chained_substitute(
    expr: SymType, *old_and_new: tuple[dict[str, SymType], dict[str, SymType]]
) -> Union[SymType, cs.DM]:
    """Internal utility to perform substitutions in chain."""
    OLD = []
    NEW = []
    for old, new in old_and_new:
        if old and new:
            for n, old_ in old.items():
                OLD.append(old_)
                NEW.append(new[n])
    return cs.substitute(expr, cs.vvcat(OLD), cs.vvcat(NEW))


def _cmp_key(
    sol: dict[str, Any], plugin_solver: Optional[str]
) -> tuple[bool, bool, float]:
    """Internal utility, similar to :func:`Solution.cmp_key`, but for native CasADi's
    solution dictionaries."""
    stats = sol["stats"]
    status = str(stats["return_status"])
    is_infeas = _is_infeas(status, plugin_solver) if plugin_solver is not None else None
    if is_infeas is None:
        is_infeas = "infeas" in status.lower()
    return is_infeas, not stats["success"], float(sol["f"])


def _get_kkt_stats(
    sol: dict[str, Any],
    ng: int,
    nonmasked_lbx_idx: Union[slice, npt.NDArray[np.int64]],
    nonmasked_ubx_idx: Union[slice, npt.NDArray[np.int64]],
    lbx: npt.NDArray[np.float64],
    ubx: npt.NDArray[np.float64],
    dlagrangian: cs.Function,
    tol_eq: float,
    tol_ineq: float,
    tol_comp: float,
    tol_stat: float,
) -> dict[str, Any]:
    """Internal utility to check the KKT conditions of a solution."""
    g_and_h = sol["g"]
    g = np.asarray(g_and_h[:ng, :].elements())
    h = np.asarray(g_and_h[ng:, :].elements())
    if (h > tol_ineq).any() or (np.abs(g) > tol_eq).any():
        return {"success": False, "return_status": "infeasible (primal)"}

    lam_g_and_h = sol["lam_g"]
    lam_h = np.asarray(lam_g_and_h[ng:, :].elements())
    lam_x = np.asarray(sol["lam_x"].elements())
    lam_lbx = -np.minimum(lam_x[nonmasked_lbx_idx], 0)
    lam_ubx = np.maximum(lam_x[nonmasked_ubx_idx], 0)
    x = np.asarray(sol["x"].elements())
    if (
        (np.abs(lam_lbx * (lbx - x[nonmasked_lbx_idx])) > tol_comp).any()
        or (np.abs(lam_ubx * (x[nonmasked_ubx_idx] - ubx)) > tol_comp).any()
        or (np.abs(lam_h * h) > tol_comp).any()
    ):
        return {"success": False, "return_status": "infeasible (complementary)"}

    lam_g = lam_g_and_h[:ng, :]
    dL = dlagrangian(x, sol["p"], lam_g, lam_h, lam_lbx, lam_ubx)
    dL = np.asarray(dL.elements()).flatten()
    if (np.abs(dL) > tol_stat).any():
        return {"success": False, "return_status": "non-stationarity point"}

    return {"success": True, "return_status": "suspected success"}


class MultistartNlp(Nlp[SymType], Generic[SymType]):
    """Base class for NLP with multistarting. This class lays the foundation for solving
    an NLP problem (described as an instance of :class:`csnlp.Nlp`) multiple times with
    different initial conditions, and requires subclasses to implement the actual
    multistarting logic in :meth:`solve_multi`.

    Parameters
    ----------
    args, kwargs
        See inherited :meth:`csnlp.Nlp.__init__`.
    starts : int
        A positive integer for the number of multiple starting guesses to optimize.

    Raises
    ------
    ValueError
        Raises if the scenario number is invalid.
    """

    is_multi: ClassVar[bool] = True
    """Flag to indicate that this is a multistart NLP."""

    def __init__(self, *args: Any, starts: int, **kwargs: Any) -> None:
        if starts <= 0:
            raise ValueError("Number of scenarios must be positive and > 0.")
        super().__init__(*args, **kwargs)
        self._starts = starts

    def __call__(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        **kwargs: Any,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        # call solve_multi only if either pars or vals0 is an iterable; otherwise, run
        # the single, base NLP
        if (pars is None or isinstance(pars, dict)) and (
            vals0 is None or isinstance(vals0, dict)
        ):
            return self.solve(pars, vals0)
        return self.solve_multi(pars, vals0, **kwargs)

    @property
    def starts(self) -> int:
        """Gets the number of starts."""
        return self._starts

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        **_: Any,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        """Solves the NLP with multiple initial conditions.

        Parameters
        ----------
        pars : dict of (str, array_like) or iterable of, optional
            An iterable that, for each multistart, contains a dictionary with, for each
            parameter in the NLP scheme, the corresponding numerical value. In case a
            single dict is passed, the same is used across all scenarions. Can be
            ``None`` if no parameters are present.
        vals0 : dict of (str, array_like) or iterable of, optional
            An iterable that, for each multistart, contains a dictionary with, for each
            variable in the NLP scheme, the corresponding initial guess. In case a
            single dict is passed, the same is used across all scenarions. By default
            ``None``, in which case  initial guesses are not passed to the solver.
        return_all_sols : bool, optional
            If ``True``, returns the solution of each multistart of the NLP; otherwise,
            only the best solution is returned. By default, ``False``.

        Returns
        -------
        Solution or list of Solutions
            Depending on the flags ``return_all_sols``, returns

            - the best solution out of all multiple starts
            - all the solutions (one per start).
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement `solve_multi`"
        )


class StackedMultistartNlp(MultistartNlp[SymType], Generic[SymType]):
    """A class that models and solves an NLP problem from multiple starting initial
    guesses by automatically stacking the original problem multiple independent times
    in the same, larger-scale NLP. This allows to solve the original problem multiple
    times via a single call to the solver."""

    def __init__(self, *args: Any, starts: int, **kwargs: Any) -> None:
        # this class essentially is a facade that hides an internal nlp in which the
        # problem (variables, parameters, etc.) are duplicated by the requested number
        # of multiple starts. For this reason, all methods are overridden to create
        # multiples of these in the hidden nlp.
        super().__init__(*args, starts=starts, **kwargs)
        self._stacked_nlp = Nlp(*args, **kwargs)  # actual nlp

    @lru_cache(maxsize=32)
    def _vars_i(self, i: int) -> dict[str, SymType]:
        """Internal utility to retrieve the variables of the ``i``-th scenario."""
        nlp = self._stacked_nlp.unwrapped
        return {k: nlp._vars[_n(k, i)] for k in self._vars}

    @lru_cache(maxsize=32)
    def _pars_i(self, i: int) -> dict[str, SymType]:
        """Internal utility to retrieve the parameters of the ``i``-th scenario."""
        nlp = self._stacked_nlp.unwrapped
        return {k: nlp._pars[_n(k, i)] for k in self._pars}

    @lru_cache(maxsize=32)
    def _dual_vars_i(self, i: int) -> dict[str, SymType]:
        """Internal utility to retrieve the dual variables of the ``i``-th scenario."""
        nlp = self._stacked_nlp.unwrapped
        return {k: nlp._dual_vars[_n(k, i)] for k in self._dual_vars}

    @invalidate_cache(_pars_i)
    def parameter(self, name: str, shape: tuple[int, int] = (1, 1)) -> SymType:
        out = super().parameter(name, shape)
        for i in range(self._starts):
            self._stacked_nlp.parameter(_n(name, i), shape)
        return out

    @invalidate_cache(_vars_i, _dual_vars_i)
    def variable(
        self,
        name: str,
        shape: tuple[int, int] = (1, 1),
        discrete: bool = False,
        lb: Union[npt.ArrayLike, cs.DM] = -np.inf,
        ub: Union[npt.ArrayLike, cs.DM] = +np.inf,
    ) -> tuple[SymType, SymType, SymType]:
        out = super().variable(name, shape, discrete, lb, ub)
        for i in range(self._starts):
            self._stacked_nlp.variable(_n(name, i), shape, discrete, lb, ub)
        return out

    @invalidate_cache(_vars_i, _dual_vars_i)
    def constraint(
        self,
        name: str,
        lhs: Union[SymType, np.ndarray, cs.DM],
        op: Literal["==", ">=", "<="],
        rhs: Union[SymType, np.ndarray, cs.DM],
        soft: bool = False,
        simplify: bool = True,
    ) -> tuple[SymType, ...]:
        expr = lhs - rhs
        if simplify:
            expr = cs.cse(cs.simplify(expr))
        out = super().constraint(name, expr, op, 0, soft, False)

        # NOTE: the line above already created the slack variables in each scenario, so
        # below we have to pass both soft=False and the expression with slack included.
        vars = self.variables
        pars = self.parameters
        expr_ = out[0]  # slack-relaxed expression in the form h(x,p)<=0 or g(x,p)==0
        op_: Literal["==", "<="] = "==" if op == "==" else "<="
        for i in range(self._starts):
            expr_i = _chained_substitute(
                expr_, (vars, self._vars_i(i)), (pars, self._pars_i(i))
            )
            self._stacked_nlp.constraint(_n(name, i), expr_i, op_, 0, False, False)
        return out

    def minimize(self, objective: SymType) -> None:
        out = super().minimize(objective)

        vars = self.variables
        pars = self.parameters
        self._fs: list[SymType] = [
            _chained_substitute(
                objective, (vars, self._vars_i(i)), (pars, self._pars_i(i))
            )
            for i in range(self._starts)
        ]
        self._stacked_nlp.minimize(sum(self._fs))
        return out

    def init_solver(
        self,
        opts: Optional[dict[str, Any]] = None,
        solver: str = "ipopt",
        type: Optional[Literal["nlp", "conic"]] = None,
    ) -> None:
        out = super().init_solver(opts, solver, type)
        self._stacked_nlp.init_solver(opts, solver, type)
        return out

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        return_stacked_sol: bool = False,
        **_: Any,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        assert not (return_stacked_sol and return_all_sols), (
            "`return_all_sols` and `return_stacked_sol` can't be both true."
        )
        if pars is not None:
            pars_iter = repeat(pars, self.starts) if isinstance(pars, dict) else pars
            pars = {_n(n, i): ps[n] for i, ps in enumerate(pars_iter) for n in ps}
        if vals0 is not None:
            v0_iter = repeat(vals0, self.starts) if isinstance(vals0, dict) else vals0
            vals0 = {_n(n, i): v0[n] for i, v0 in enumerate(v0_iter) for n in v0}
        multi_sol = self._stacked_nlp.solve(pars, vals0)
        if return_stacked_sol:
            return multi_sol

        x_ = self._x
        p_ = self._p
        lam_g_and_h_ = cs.vertcat(self._lam_g, self._lam_h)
        lam_lbx_and_ubx_ = cs.vertcat(self._lam_lbx, self._lam_ubx)
        vars_ = self.variables
        pars_ = self.parameters
        duals_ = self.dual_variables
        fs = [float(multi_sol.value(f)) for f in self._fs]

        all_vars = cs.vertcat(x_, lam_g_and_h_, lam_lbx_and_ubx_, p_)
        splits = np.cumsum(
            (0, x_.size1(), lam_g_and_h_.size1(), lam_lbx_and_ubx_.size1(), p_.size1())
        )
        solver_plugin = self.unwrapped._solver_plugin

        def get_ith_sol(idx: int) -> EagerSolution[SymType]:
            vals = {n: multi_sol.vals[_n(n, idx)] for n in vars_}
            dual_vals = {n: multi_sol.dual_vals[_n(n, idx)] for n in duals_}
            # get the value of p, and, lam g, lam h and lam lbx and lam ubx in a single
            # substitution to save some time
            all_vals = multi_sol.value(
                _chained_substitute(
                    all_vars,
                    (vars_, self._vars_i(idx)),
                    (pars_, self._pars_i(idx)),
                    (duals_, self._dual_vars_i(idx)),
                )
            )
            x, lam_g_and_h, lam_lbx_and_ubx, p = cs.vertsplit(all_vals, splits)
            return EagerSolution(
                fs[idx],
                p_,
                p,
                x_,
                x,
                lam_g_and_h_,
                lam_g_and_h,
                lam_lbx_and_ubx_,
                lam_lbx_and_ubx,
                vars_,
                vals,
                duals_,
                dual_vals,
                multi_sol.stats,
                solver_plugin,
            )

        if return_all_sols:
            return [get_ith_sol(i) for i in range(self._starts)]
        return get_ith_sol(np.argmin(fs).item())

    def remove_variable_bounds(
        self,
        name: str,
        direction: Literal["lb", "ub", "both"],
        idx: Union[None, tuple[int, int], list[tuple[int, int]]] = None,
    ) -> None:
        idx = [idx] if isinstance(idx, tuple) else list(idx)
        super().remove_variable_bounds(name, direction, idx)
        for i in range(self._starts):
            self._stacked_nlp.remove_variable_bounds(_n(name, i), direction, idx)

    def remove_constraints(
        self, name: str, idx: Union[None, tuple[int, int], list[tuple[int, int]]] = None
    ) -> None:
        idx = [idx] if isinstance(idx, tuple) else list(idx)
        super().remove_constraints(name, idx)
        for i in range(self._starts):
            self._stacked_nlp.remove_constraints(_n(name, i), idx)


class ParallelMultistartNlp(MultistartNlp[SymType], Generic[SymType]):
    """A class that solves an NLP problem multiple times, with different initial
    starting conditions, via parallelization of the computations via :mod:`joblib`.

    Parameters
    ----------
    args, kwargs
        See inherited :meth:`csnlp.Nlp.__init__`.
    starts : int
        A positive integer for the number of multiple starting guesses to optimize.
    parallel_kwargs: dict, optional
        A dictionary with keyword arguments to be used to instantiate the parallel
        :class:`joblib.Parallel` backend. By default, ``None``, in which case the
        parallel backend is instantiated with default arguments.

    Raises
    ------
    ValueError
        Raises if the scenario number is invalid.
    """

    def __init__(
        self,
        *args: Any,
        starts: int,
        parallel_kwargs: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, starts=starts, **kwargs)
        self._parallel_kwargs = parallel_kwargs if parallel_kwargs is not None else {}
        self._parallel: Optional[Parallel] = None

    def initialize_parallel(self) -> None:
        """Initializes the parallel backend."""
        self._parallel = Parallel(**self._parallel_kwargs)
        self._parallel.__enter__()

    def terminate_parallel(self) -> None:
        """Terminates the parallel backend."""
        self._parallel.__exit__(None, None, None)

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        **_: Any,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        if self._solver is None:
            raise RuntimeError("Solver uninitialized.")
        if self._parallel is None:
            self.initialize_parallel()
        shared_kwargs = {
            "lbx": self._lbx.data,
            "ubx": self._ubx.data,
            "lbg": np.concatenate((np.zeros(self.ng), np.full(self.nh, -np.inf))),
            "ubg": 0,
        }
        pars_iter = (
            repeat(pars, self.starts)
            if pars is None or isinstance(pars, dict)
            else pars
        )
        vals0_iter = (
            repeat(vals0, self.starts)
            if vals0 is None or isinstance(vals0, dict)
            else vals0
        )
        kwargs = (
            self._process_pars_and_vals0(shared_kwargs.copy(), p, v0)
            for p, v0 in zip(pars_iter, vals0_iter)
        )
        sols: Iterator[dict[str, Any]] = self._parallel(
            delayed(_solve_and_get_stats)(self._solver, kw) for kw in kwargs
        )
        if return_all_sols:
            return [LazySolution.from_casadi_solution(sol, self) for sol in sols]
        best_sol = min(sols, key=lambda s: _cmp_key(s, self._solver_plugin))
        return LazySolution.from_casadi_solution(best_sol, self)


class MappedMultistartNlp(MultistartNlp[SymType], Generic[SymType]):
    """A class that solves an NLP problem multiple times, with different initial
    conditions, in parallel via :func:`casadi.Function.map` parallelization.

    See
    `this wiki <https://github.com/casadi/casadi/wiki/FAQ:-How-to-use-map%28%29-and-mapaccum%28%29-to-speed-up-calculations%3F>`_
    for more details on how to enable parallelization via mapping in CasADi.

    Parameters
    ----------
    args, kwargs
        See inherited :meth:`csnlp.Nlp.__init__`.
    starts : int
        A positive integer for the number of multiple starting guesses to optimize.
    parallelization : "serial", "unroll", "inline", "thread", "openmp"
        The type of parallelization to use (see :func:`casadi.Function.map`). By
        default, ``"serial"`` is selected.
    max_num_threads : int, optional
        Maximum number of threads to use in parallelization; if ``None``, the number of
        threads is equal to the number of starts.

    Raises
    ------
    ValueError
        Raises if the scenario number is invalid.
    """

    def __init__(
        self,
        *args: Any,
        starts: int,
        parallelization: Literal[
            "serial", "unroll", "inline", "thread", "openmp"
        ] = "serial",
        max_num_threads: Optional[int] = None,
        tol_stat: float = 1e-8,
        tol_eq: float = 1e-8,
        tol_ineq: float = 1e-8,
        tol_comp: float = 1e-8,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, starts=starts, **kwargs)
        self._mapped_solver: Optional[MemorizedFunc] = None
        self._parallelization = parallelization
        self._max_num_threads = max_num_threads
        self._tol_eq = tol_eq
        self._tol_ineq = tol_ineq
        self._tol_comp = tol_comp
        self._tol_stat = tol_stat

    def init_solver(self, *args: Any, **kwargs: Any) -> None:
        super().init_solver(*args, **kwargs)
        solver: cs.Function = self._solver.func
        mapped_solver = solver.map(
            self._starts, self._parallelization, self._max_num_threads or self._starts
        )
        self._mapped_solver = self._cache.cache(mapped_solver)

        lagrangian = (
            self.f
            + cs.dot(self.lam_g, self.g)
            + cs.dot(self.lam_h, self.h)
            + cs.dot(self.lam_lbx, self.h_lbx)
            + cs.dot(self.lam_ubx, self.h_ubx)
        )
        dL = cs.jacobian(lagrangian, self.x).T
        self._kkt_stationarity = cs.Function(
            "kkt_stationarity",
            [self.x, self.p, self.lam_g, self.lam_h, self.lam_lbx, self.lam_ubx],
            [dL],
            ["x", "p", "lam_g", "lam_h", "lam_lbx", "lam_ubx"],
            ["dL"],
        )

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        _return_mapped_sol: bool = False,
        **_: Any,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        assert not (_return_mapped_sol and return_all_sols), (
            "`return_all_sols` and `_return_mapped_sol` can't be both true."
        )
        if self._mapped_solver is None:
            raise RuntimeError("Solver uninitialized.")
        pars_iter = (
            repeat(pars, self.starts)
            if pars is None or isinstance(pars, dict)
            else pars
        )
        vals0_iter = (
            repeat(vals0, self.starts)
            if vals0 is None or isinstance(vals0, dict)
            else vals0
        )
        x0s = []
        ps = []
        default = cs.DM.zeros(0, 1)
        for p, v0 in zip(pars_iter, vals0_iter):
            kwargs = self._process_pars_and_vals0({}, p, v0)
            x0s.append(kwargs.get("x0", default))
            ps.append(kwargs.get("p", default))
        single_kwargs = {
            "x0": cs.hcat(x0s),
            "p": cs.hcat(ps),
            "lbx": self._lbx.data,
            "ubx": self._ubx.data,
            "lbg": np.concatenate((np.zeros(self.ng), np.full(self.nh, -np.inf))),
            "ubg": 0,
        }
        single_sol: dict[str, cs.DM] = self._mapped_solver(**single_kwargs)

        # if the user wants the mapped solution, return it, though it's not a Solution
        # and it's mostly for debugging
        if _return_mapped_sol:
            single_sol["p"] = single_kwargs["p"]
            single_sol["stats"] = self._mapped_solver.func.stats()
            return single_sol  # NOTE: this is NOT a Solution object

        sols: list[dict[str, Any]] = []
        ng = self.ng
        nonmasked_lbx_idx = self.nonmasked_lbx_idx
        nonmasked_ubx_idx = self.nonmasked_ubx_idx
        lbx = self._lbx.data[nonmasked_lbx_idx]
        ubx = self._ubx.data[nonmasked_ubx_idx]
        dlagrangian = self._kkt_stationarity
        tol_eq = self._tol_eq
        tol_ineq = self._tol_ineq
        tol_comp = self._tol_comp
        tol_stat = self._tol_stat
        for i, p in enumerate(ps):
            sol_i = {k: v[:, i] if v.size2() else v for k, v in single_sol.items()}
            sol_i["p"] = p
            sol_i["stats"] = _get_kkt_stats(
                sol_i,
                ng,
                nonmasked_lbx_idx,
                nonmasked_ubx_idx,
                lbx,
                ubx,
                dlagrangian,
                tol_eq,
                tol_ineq,
                tol_comp,
                tol_stat,
            )
            sols.append(sol_i)
        if return_all_sols:
            return [LazySolution.from_casadi_solution(sol, self) for sol in sols]
        # NOTE: since the return status is here hand-crafted, don't pass the plugin name
        # for detecting infeasibility so as to default to searching for "infeas"
        best_sol = min(sols, key=lambda s: _cmp_key(s, None))
        return LazySolution.from_casadi_solution(best_sol, self)



"""A module for defining NLPs and its building blocks. From simplest to most complex,
and following the inheritance hierarchy, these are:

- :class:`csnlp.nlps.HasParameters`: a class for the creation and storage of symbolic
  parameters for NLP problems
- :class:`csnlp.nlps.HasVariables`: a class for the creation and storage of symbolic
  variables for NLP problems
- :class:`csnlp.nlps.HasConstraints`: a class for the creation and storage of symbolic
  constraints (dependent on variables and parameters) for NLP problems. The
  constraints can be equality, inequality, lower- or upper-bound.
- :class:`csnlp.nlps.HasObjective`: a class for the assignment of a scalar minimization
  objective function (dependent on variables and parameters).
- :class:`csnlp.Nlp`: a class that combines all the above building blocks into a
  full-fledged NLP problem.
"""

__all__ = ["HasConstraints", "HasObjective", "HasParameters", "HasVariables", "Nlp"]

from .constraints import HasConstraints
from .nlp import Nlp
from .objective import HasObjective
from .parameters import HasParameters
from .variables import HasVariables



"""A collection of functions for manipulating data in CasADi, in particular, on how
to convert to and from numpy arrays and CasADi symbolic variables, and how to find
the indices of a symbolic variable in a vector of symbols."""

from typing import Union

import casadi as cs
import numpy as np
import numpy.typing as npt


def array2cs(x: np.ndarray) -> Union[cs.SX, cs.MX]:
    """Converts the numpy array ``x`` containing scalar symbolic variables to a single
    symbolic variable (see :func:`cs2array` for the opposite functionality).

    Parameters
    ----------
    x : numpy array of casadi.SX or MX
        Array whose entries are either MX or SX.

    Returns
    -------
    casadi.SX or MX
        A single SX or MX instance whose entries are the ``x``'s entries.

    Raises
    ------
    ValueError
        Raises if the array is empty (zero dimensions), or if it has more than 2
        dimensions.

    Notes
    -----
    Note that all entries in ``x`` must have the same type, either SX or MX.
    """
    if isinstance(x, (cs.SX, cs.MX, cs.DM)):
        return x
    cls = type(next(x.flat))
    if x.dtype != object or cls is cs.DM:
        return cs.DM(x)  # meaning it is not SX or MX, thus a number
    ndim = x.ndim
    if ndim == 0:
        return x.item()
    if ndim == 1:
        numel = x.size
        shape = (numel, 1)
        indices = range(numel)
    elif ndim == 2:
        shape = x.shape
        indices = np.ndindex(shape)
    else:
        raise ValueError("Can only convert 1D and 2D arrays to CasADi SX or MX.")
    m: Union[cs.SX, cs.MX] = cls(*shape)
    for idx in indices:
        m[idx] = x[idx]
    return m


def cs2array(x: Union[cs.MX, cs.SX]) -> np.ndarray:
    """Converts the symbolic variable ``x`` to a numpy array of scalar symbolic
    variables (see :func:`array2cs` for the opposite functionality).

    Inspired by the implementation in
    `MPCTools <https://bitbucket.org/rawlings-group/mpc-tools-casadi/src/master/mpctools/tools.py>`_.

    Parameters
    ----------
    x : casadi.SX or MX
        A symbolic variable (with multiple entries).

    Returns
    -------
    numpy array of casadi.SX or MX
        The array containing the symbolic variable scalars.
    """
    if isinstance(x, np.ndarray):
        return x
    if isinstance(x, cs.DM):
        return x.toarray()

    shape = x.shape
    y = np.empty(shape, object)
    for idx in np.ndindex(shape):
        y[idx] = x[idx]
    return y


def find_index_in_vector(
    V: Union[cs.SX, cs.MX], a: Union[cs.SX, cs.MX]
) -> npt.NDArray[np.int_]:
    """Finds the indices of ``a`` in ``V``.

    Parameters
    ----------
    V : casadi.SX or MX
        The vector in which to search.
    a : casadi.SX or MX
        The vector to search for.

    Returns
    -------
    array of integers
        The indices of ``a`` in ``V``.

    Raises
    ------
    ValueError
        Raises if ``V`` or ``a`` are not vectors.
    """
    if not V.is_vector() or not a.is_vector():
        raise ValueError("`V` and `a` must be vectors.")

    sp: cs.Sparsity = cs.jacobian_sparsity(a, V)
    idx = np.asarray(sp.get_crs()[1], int)
    if idx.size != a.numel():
        raise RuntimeError("invalid subset: some entries of `a` were not found in `V`")
    return idx



import math as _math
import os
import pickle
import tempfile
import unittest
from copy import deepcopy
from itertools import product
from typing import Any, Optional

import casadi as cs
import cvxpy as cp
import numpy as np
from parameterized import parameterized
from scipy.special import digamma, gammaln
from scipy.stats import norm

from csnlp import Nlp
from csnlp.core.solutions import subsevalf
from csnlp.util import docs, io, math

TMPFILENAME: str = ""


class DictClass:
    def __init__(self) -> None:
        super().__init__()
        self.z, self.w = 3, 4
        self.sym2 = cs.SX.sym("y", 3, 1)


class TestIo(unittest.TestCase):
    @parameterized.expand(
        [
            ("pkl",),
            ("xz",),
            ("pbz2",),
            ("gz",),
            ("bt",),
            ("bl2",),
            ("mat",),
            ("npz",),
        ]
    )
    def test_save_and_load__preserve_data_correctly(self, ext: str):
        global TMPFILENAME
        TMPFILENAME = f"{next(tempfile._get_candidate_names())}.{ext}"
        data = {"x": 5, "y": "ciao", "w": {"ci": "ao"}}
        io.save(TMPFILENAME, **data)
        data2 = io.load(TMPFILENAME)
        self.assertDictEqual(data, data2)

    def test_save_and_load__one_key_dict_is_simplified(self):
        global TMPFILENAME
        TMPFILENAME = f"{next(tempfile._get_candidate_names())}.pkl"
        data = {"x": 5}
        io.save(TMPFILENAME, **data)
        data2 = io.load(TMPFILENAME)
        self.assertIsNot(data2, dict)
        self.assertEqual(data["x"], data2)

    def tearDown(self) -> None:
        try:
            os.remove(TMPFILENAME)
        finally:
            return super().tearDown()

    @parameterized.expand([(False,), (True,)])
    def test_is_pickleable_and_deepcopy_able(self, copy: bool):
        this = DictClass()
        if copy:
            other = deepcopy(this)
        else:
            with cs.global_pickle_context():
                pickled = pickle.dumps(this)
            with cs.global_unpickle_context():
                other = pickle.loads(pickled)

        self.assertIsNot(this, other)
        self.assertEqual(this.z, other.z)
        self.assertEqual(this.w, other.w)
        if copy:
            self.assertTrue(cs.is_equal(this.sym2, other.sym2))
        else:
            self.assertTupleEqual(this.sym2.shape, other.sym2.shape)


class TestMath(unittest.TestCase):
    def test_log(self):
        base = np.random.rand() * 10
        x = np.random.rand() * 10
        self.assertEqual(math.log(x), _math.log(x))
        self.assertEqual(math.log(x, base), _math.log(x, base))

    @parameterized.expand(product([(3, 1), (1, 3), (3, 4)], [-2, -1, 0, 1, None]))
    def test_prod(self, shape: tuple[int, int], axis: Optional[int]):
        x = np.random.randn(*shape) * 3
        p = np.prod(x, axis=axis, keepdims=True)
        x_sx = cs.SX.sym("X", *shape)
        p_sx = subsevalf(math.prod(x_sx, axis=axis), x_sx, x)
        x_mx = cs.MX.sym("X", *shape)
        p_mx = subsevalf(math.prod(x_mx, axis=axis), x_mx, x)
        x_dm = cs.DM(x)
        p_dm = math.prod(x_dm, axis=axis)
        np.testing.assert_allclose(p, p_sx)
        np.testing.assert_allclose(p, p_mx)
        np.testing.assert_allclose(p, p_dm)

    def test_norm_cdf(self):
        shape = (3, 4)
        x_sx = cs.SX.sym("X", *shape)
        loc_sx = cs.SX.sym("loc", *shape)
        scale_sx = cs.SX.sym("scale", *shape)
        x_mx = cs.MX.sym("X", *shape)
        loc_mx = cs.MX.sym("loc", *shape)
        scale_mx = cs.MX.sym("scale", *shape)
        x = np.random.randn(*shape)
        loc = np.random.randn(*shape)
        scale = np.random.rand(*shape)
        cdf_sx = subsevalf(
            math.normal_cdf(x_sx, loc=loc_sx, scale=scale_sx),
            [x_sx, loc_sx, scale_sx],
            [x, loc, scale],
        )
        cdf_mx = subsevalf(
            math.normal_cdf(x_mx, loc=loc_mx, scale=scale_mx),
            [x_mx, loc_mx, scale_mx],
            [x, loc, scale],
        )
        cdf = norm.cdf(x, loc=loc, scale=scale)

        np.testing.assert_allclose(cdf, cdf_sx, atol=1e-7, rtol=1e-5)
        np.testing.assert_allclose(cdf, cdf_mx, atol=1e-7, rtol=1e-5)

    def test_norm_ppf(self):
        shape = (3, 4)
        x_sx = cs.SX.sym("X", *shape)
        loc_sx = cs.SX.sym("loc", *shape)
        scale_sx = cs.SX.sym("scale", *shape)
        x_mx = cs.MX.sym("X", *shape)
        loc_mx = cs.MX.sym("loc", *shape)
        scale_mx = cs.MX.sym("scale", *shape)
        x = np.random.rand(*shape)
        loc = np.random.randn(*shape)
        scale = np.random.rand(*shape)
        cdf_sx = subsevalf(
            math.normal_ppf(x_sx, loc=loc_sx, scale=scale_sx),
            [x_sx, loc_sx, scale_sx],
            [x, loc, scale],
        )
        cdf_mx = subsevalf(
            math.normal_ppf(x_mx, loc=loc_mx, scale=scale_mx),
            [x_mx, loc_mx, scale_mx],
            [x, loc, scale],
        )
        cdf = norm.ppf(x, loc=loc, scale=scale)

        np.testing.assert_allclose(cdf, cdf_sx, atol=1e-5, rtol=1e-5)
        np.testing.assert_allclose(cdf, cdf_mx, atol=1e-5, rtol=1e-5)

    @parameterized.expand(
        [
            ((3, 4), cs.DM([3, 3, 3, 3])),
            (([1, 2, 3, 4], 3), cs.DM(sum(([i] * 3 for i in range(1, 5)), []))),
            ((cs.DM([[1, 2], [3, 4]]), (1, 2)), cs.DM([[1, 1, 2, 2], [3, 3, 4, 4]])),
        ]
    )
    def test_repeat(self, inputs, expected):
        actual = math.repeat(*inputs)
        np.testing.assert_array_equal(actual, expected)

    @parameterized.expand(product(("SX", "MX"), (1, "inf")))
    def test_norm_1_and_inf(self, sym_type, ord):
        # see Boyd & Vandenberghe, ex. 4.11 (a) and (b)
        # https://egrcc.github.io/docs/math/cvxbook-solutions.pdf
        m, n = sorted(np.random.randint(10, 20, size=2))
        A = np.random.randn(n, m)
        b = np.random.randn(n)

        x = cp.Variable(m)
        objective = cp.Minimize(cp.norm(A @ x - b, ord))
        problem = cp.Problem(objective)
        problem.solve()
        f_cp = problem.value
        x_cp = x.value

        nlp = Nlp(sym_type)
        x, _, _ = nlp.variable("x", (m, 1))
        norm_fun = math.norm_1 if ord == 1 else math.norm_inf
        nlp.minimize(norm_fun(nlp, "", A @ x - b))
        nlp.init_solver(solver="clp")
        sol = nlp.solve()
        f_csnlp = sol.f
        x_csnlp = sol.value(x).toarray().flatten()

        np.testing.assert_allclose(f_cp, f_csnlp, rtol=1e-6, atol=1e-6)
        np.testing.assert_allclose(x_cp, x_csnlp, rtol=1e-6, atol=1e-6)

    @parameterized.expand(
        [
            (
                5,
                7,
                [
                    1.000000000190015,
                    76.18009172947146,
                    -86.50532032941677,
                    24.01409824083091,
                    -1.231739572450155,
                    0.1208650973866179e-2,
                    -0.5395239384953e-5,
                ],
            ),
            (
                7,
                9,
                [
                    0.99999999999980993227684700473478,
                    676.520368121885098567009190444019,
                    -1259.13921672240287047156078755283,
                    771.3234287776530788486528258894,
                    -176.61502916214059906584551354,
                    12.507343278686904814458936853,
                    -0.13857109526572011689554707,
                    9.984369578019570859563e-6,
                    1.50563273514931155834e-7,
                ],
            ),
            (
                8,
                12,
                [
                    0.9999999999999999298,
                    1975.3739023578852322,
                    -4397.3823927922428918,
                    3462.6328459862717019,
                    -1156.9851431631167820,
                    154.53815050252775060,
                    -6.2536716123689161798,
                    0.034642762454736807441,
                    -7.4776171974442977377e-7,
                    6.3041253821852264261e-8,
                    -2.7405717035683877489e-8,
                    4.0486948817567609101e-9,
                ],
            ),
            (
                4.7421875,
                15,
                [
                    0.99999999999999709182,
                    57.156235665862923517,
                    -59.597960355475491248,
                    14.136097974741747174,
                    -0.49191381609762019978,
                    0.33994649984811888699e-4,
                    0.46523628927048575665e-4,
                    -0.98374475304879564677e-4,
                    0.15808870322491248884e-3,
                    -0.21026444172410488319e-3,
                    0.21743961811521264320e-3,
                    -0.16431810653676389022e-3,
                    0.84418223983852743293e-4,
                    -0.26190838401581408670e-4,
                    0.36899182659531622704e-5,
                ],
            ),
        ]
    )
    def test_godfrey_coefficients(self, g, n, expected):
        actual = math.godfrey_coefficients(g, n).astype(float)
        np.testing.assert_allclose(actual, expected, rtol=1e-6, atol=1e-6)

    def test_gammaln(self):
        z = cs.SX.sym("z")
        y = math.gammaln(z, 9, 12)
        cs_gammaln = cs.Function("gammaln", [z], [y])

        z = np.linspace(0.0, 10, 10000)
        expected = gammaln(z)
        actual = cs_gammaln(z).full().flatten()
        np.testing.assert_allclose(actual, expected, rtol=1e-6, atol=1e-6)

    def test_digamma1p(self):
        z = cs.SX.sym("z")
        y = math.digamma1p(z, 3)
        cs_digamma1p = cs.Function("digamma1p", [z], [y])

        z = np.linspace(1e-6, 10, 10_000)
        expected = digamma(z + 1)
        actual = cs_digamma1p(z).full().flatten()
        np.testing.assert_allclose(actual, expected, rtol=1e-2, atol=1e-3)

    def test_digamma(self):
        z = cs.SX.sym("z")
        y = math.digamma(z, 3)
        cs_digamma = cs.Function("digamma", [z], [y])

        z = np.linspace(1e-6, 10, 10_000)
        expected = digamma(z)
        actual = cs_digamma(z).full().flatten()
        np.testing.assert_allclose(actual, expected, rtol=1e-4, atol=1e-4)


class TestDocs(unittest.TestCase):
    def test_get_casadi_plugins(self):
        all_plugins = docs.get_casadi_plugins()
        self.assertIsInstance(all_plugins, dict)
        self.assertGreater(len(all_plugins), 0)
        for name, plugins in all_plugins.items():
            self.assertIsInstance(name, str)
            self.assertIsInstance(plugins, list)
            for p in plugins:
                self.assertIsInstance(p, str)

    def test_list_available_solvers(self):
        solvers = docs.list_available_solvers()
        self.assertIsInstance(solvers, dict)
        self.assertIn("nlp", solvers)
        self.assertIn("qp", solvers)
        for key in ("nlp", "qp"):
            self.assertIsInstance(solvers[key], list)
            for solver in solvers[key]:
                self.assertIsInstance(solver, str)
                self.assertGreater(len(solver), 0)

    def test_get_solver_options(self):
        options = docs.get_solver_options("ipopt", False)
        self.assertIsInstance(options, dict)


if __name__ == "__main__":
    unittest.main()



import pickle
import unittest
from copy import deepcopy
from itertools import product
from typing import Union
from unittest.mock import Mock

import casadi as cs
import numpy as np
from parameterized import parameterized, parameterized_class

from csnlp import Nlp
from csnlp.core.solutions import subsevalf
from csnlp.util.math import log

OPTS = {
    "expand": True,
    "print_time": False,
    "ipopt": {
        "max_iter": 500,
        "sb": "yes",
        # for debugging
        "print_level": 0,
        "print_user_options": "no",
        "print_options_documentation": "no",
    },
}


@parameterized_class(["sym_type"], [("SX",), ("MX",)])
class TestNlp(unittest.TestCase):
    def cmp(
        self,
        lhs: Union[cs.SX, cs.MX],
        rhs: Union[cs.SX, cs.MX],
        vars: Union[list[cs.SX], list[cs.MX]],
        force_numerical: bool = False,
    ) -> bool:
        if isinstance(rhs, (int, float)):
            rhs = np.full(lhs.shape, rhs)
        if not force_numerical and self.sym_type == "SX":
            return cs.is_equal(lhs, rhs)
        old = cs.vvcat(vars)
        new = np.random.randn(*old.shape)
        np.testing.assert_allclose(subsevalf(lhs, old, new), subsevalf(rhs, old, new))
        return True

    def test_init__raises__with_invalid_sym_type(self):
        with self.assertRaises(AttributeError):
            Nlp(sym_type="a_random_sym_type")

    def test_init__saves_sym_type(self):
        nlp = Nlp(sym_type=self.sym_type)
        self.assertEqual(nlp._sym_type.__name__, self.sym_type)

    def test_parameter__creates_correct_parameter(self):
        shape1 = (4, 3)
        shape2 = (2, 2)
        _np = np.prod(shape1) + np.prod(shape2)
        nlp = Nlp(sym_type=self.sym_type, debug=True)
        p1 = nlp.parameter("p1", shape1)
        p2 = nlp.parameter("p2", shape2)
        self.assertEqual(p1.shape, shape1)
        self.assertEqual(p2.shape, shape2)
        self.assertEqual(nlp.np, _np)

        p = cs.vertcat(cs.vec(p1), cs.vec(p2))
        self.cmp(nlp.p, p, vars=[p1, p2])

        i = 0
        for name, shape in [("p1", shape1), ("p2", shape2)]:
            for _ in range(np.prod(shape)):
                self.assertEqual(name, nlp.debug.p_describe(i).name)
                self.assertEqual(shape, nlp.debug.p_describe(i).shape)
                i += 1
        with self.assertRaises(IndexError):
            nlp.debug.p_describe(_np + 1)

        self.assertTrue(cs.is_equal(nlp.parameters["p1"], p1))
        self.assertTrue(cs.is_equal(nlp.parameters["p2"], p2))

    def test_parameter__raises__with_parameters_with_same_name(self):
        nlp = Nlp(sym_type=self.sym_type)
        nlp.parameter("p")
        with self.assertRaises(ValueError):
            nlp.parameter("p")

    def test_variable__discrete__is_invalidated_when_new_variable_is_added(self):
        n_variables = 30
        np_random = np.random.default_rng()

        nlp = Nlp(sym_type=self.sym_type)
        discretes = np_random.random(n_variables) < 0.5
        shapes = [tuple(np_random.integers(1, 10, 2)) for _ in range(n_variables)]
        discrete_dict = {f"x{i}": d for i, d in enumerate(discretes)}
        discrete = np.concatenate(
            [
                (np.ones if d else np.zeros)(np.prod(s), dtype=bool)
                for s, d in zip(shapes, discrete_dict.values())
            ]
        )
        cumnx = np.cumsum([np.prod(s) for s in shapes])

        for i, (shape, discrete_) in enumerate(zip(shapes, discretes)):
            nlp.variable(f"x{i}", shape, discrete_)
            np.testing.assert_array_equal(nlp.discrete, discrete[: cumnx[i]])

    def test_variable__creates_correct_variable(self):
        n_variables = 30
        np_random = np.random.default_rng()

        nlp = Nlp(sym_type=self.sym_type, debug=True)
        discretes = np_random.random(n_variables) < 0.5
        shapes = [tuple(np_random.integers(1, 10, 2)) for _ in range(n_variables)]
        lbs = [np_random.random(size=s) - 1 for s in shapes]
        ubs = [np_random.random(size=s) + 1 for s in shapes]
        xs, lams_lb, lams_ub = [], [], []
        for i, (shape, discrete, lb, ub) in enumerate(zip(shapes, discretes, lbs, ubs)):
            x, lam_lb, lam_ub = nlp.variable(f"x{i}", shape, discrete, lb, ub)
            xs.append(x)
            lams_lb.append(lam_lb)
            lams_ub.append(lam_ub)

        for shape, x, lam_lb, lam_ub in zip(shapes, xs, lams_lb, lams_ub):
            self.assertEqual(x.shape, shape)
            self.assertEqual(lam_lb.shape, (np.prod(shape), 1))
            self.assertEqual(lam_ub.shape, (np.prod(shape), 1))

        nx = sum(np.prod(s) for s in shapes)
        self.assertEqual(nlp.nx, nx)
        self.cmp(nlp.x, cs.vvcat(xs), vars=xs)

        discrete = np.concatenate(
            [
                (np.ones if d else np.zeros)(np.prod(s), dtype=bool)
                for s, d in zip(shapes, discretes)
            ]
        )
        np.testing.assert_array_equal(nlp.discrete, discrete)

        lb = cs.vvcat(lbs)
        np.testing.assert_allclose(nlp.lbx, lb.full().flat)
        ub = cs.vvcat(ubs)
        np.testing.assert_allclose(nlp.ubx, ub.full().flat)

        k = 0
        for i, shape in enumerate(shapes):
            name = f"x{i}"
            for _ in range(np.prod(shape)):
                self.assertEqual(name, nlp.debug.x_describe(k).name)
                self.assertEqual(shape, nlp.debug.x_describe(k).shape)
                k += 1
            self.assertTrue(cs.is_equal(nlp.variables[name], xs[i]))

        with self.assertRaises(IndexError):
            nlp.debug.x_describe(nx + 1)

    def test_variable__raises__with_variables_with_same_name(self):
        nlp = Nlp(sym_type=self.sym_type)
        nlp.variable("x")
        with self.assertRaises(ValueError):
            nlp.variable("x")

    def test_variable__raises__with_invalid_bounds(self):
        nlp = Nlp(sym_type=self.sym_type)
        with self.assertRaises(ValueError):
            nlp.variable("x", lb=1, ub=0)

    def test_minimize__raises__with_nonscalar_objective(self):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", (5, 1))[0]
        with self.assertRaises(ValueError):
            nlp.minimize(x)

    def test_minimize__sets_objective_correctly(self):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", (5, 1))[0]
        f = x.T @ x
        nlp.minimize(f)
        self.assertTrue(cs.is_equal(nlp.f, f))

    def test_constraint__raises__with_constraints_with_same_name(self):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        nlp.constraint("c1", x, "<=", 5)
        with self.assertRaises(ValueError):
            nlp.constraint("c1", x, "<=", 5)

    def test_constraint__raises__with_unknown_operator(self):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        for op in ["=", ">", "<"]:
            with self.assertRaises(ValueError):
                nlp.constraint("c1", x, op, 5)

    def test_constraint__raises__with_nonsymbolic_terms(self):
        nlp = Nlp(sym_type=self.sym_type)
        with self.assertRaises(TypeError):
            nlp.constraint("c1", 5, "==", 5)

    @parameterized.expand([("==", "=="), ("==", ">="), (">=", ">=")])
    def test_constraint__creates_constraint_correctly(self, op1: str, op2: str):
        shape1, shape2 = (4, 3), (2, 2)
        nc1, nc2 = np.prod(shape1), np.prod(shape2)
        nc = nc1 + nc2
        nlp = Nlp(sym_type=self.sym_type, debug=True)
        x = nlp.variable("x", shape1)[0]
        y = nlp.variable("y", shape2)[0]
        c1, lam_c1 = nlp.constraint("c1", x, op1, 5)
        c2, lam_c2 = nlp.constraint("c2", 5, op2, y)

        self.assertEqual(c1.shape, shape1)
        self.assertEqual(lam_c1.shape, (nc1, 1))
        self.assertEqual(c2.shape, shape2)
        self.assertEqual(lam_c2.shape, (nc2, 1))
        self.assertEqual(nlp.ng + nlp.nh, nc)

        i, prev_op = 0, op1
        for c, name, op in zip((c1, c2), ("c1", "c2"), (op1, op2)):
            grp = "g" if op == "==" else "h"
            if op != prev_op:
                i = 0
            describe = getattr(nlp.debug, f"{grp}_describe")
            for _ in range(np.prod(c.shape)):
                self.assertEqual(name, describe(i).name)
                self.assertEqual(c.shape, describe(i).shape)
                i += 1
            with self.assertRaises(IndexError):
                describe(nc + 1)

        c = cs.veccat(c1, c2)
        lam = cs.vertcat(lam_c1, lam_c2)
        expected_c = cs.vertcat(nlp.g, nlp.h)
        expected_lam = cs.vertcat(nlp.lam_g, nlp.lam_h)
        self.cmp(expected_c, c, vars=[x, y])
        self.cmp(expected_lam, lam, vars=[lam_c1, lam_c2])

    def test_constraint__adds_soft_variable_correctly(self):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        nlp.constraint("c0", x, ">=", 0, soft=True)
        self.assertEqual(nlp.nx, 2)
        self.assertIn("slack_c0", nlp._vars)

    def test_constraint__solves_correctly__with_soft_variable(self):
        # From https://www.gams.com/latest/docs/UG_EMP_SoftConstraints.html
        # solve with manual soft variable
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", (2, 1), lb=0)[0]
        v = nlp.variable("v", lb=0)[0]
        nlp.constraint("c0", 3 * x[0] + x[1], "<=", 5)
        nlp.constraint("c1", v, ">=", 20 - x[1] ** 2)
        nlp.minimize(-(x[0] ** 2) + 5 * (cs.log(x[0]) - 1) ** 2 + 2 * v)
        nlp.init_solver(OPTS)
        sol1 = nlp.solve(vals0={"x": [1, 1], "v": 0})

        # solve with automatic soft variable
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", (2, 1), lb=0)[0]
        nlp.constraint("c0", 3 * x[0] + x[1], "<=", 5)
        _, _, v = nlp.constraint("c1", x[1] ** 2, ">=", 20, soft=True)
        nlp.minimize(-(x[0] ** 2) + 5 * (cs.log(x[0]) - 1) ** 2 + 2 * v)
        nlp.init_solver(OPTS)
        sol2 = nlp.solve(vals0={"x": [1, 1], "slack_c1": 0})

        # assert results equal
        np.testing.assert_allclose(sol1.f, sol2.f)
        np.testing.assert_allclose(sol1.vals["x"], sol2.vals["x"])
        np.testing.assert_allclose(sol1.vals["v"], sol2.vals["slack_c1"])

    @parameterized.expand(
        [
            (True, "==", "=="),
            (True, "==", ">="),
            (True, ">=", ">="),
            (False, "==", "=="),
            (False, "==", ">="),
            (False, ">=", ">="),
        ]
    )
    def test_dual__returns_dual_variables_correctly(
        self, flag: bool, op1: str, op2: str
    ):
        shape1, shape2 = (4, 3), (2, 2)
        nlp = Nlp(sym_type=self.sym_type, remove_redundant_x_bounds=flag)
        x, lam_lb_x, lam_ub_x = nlp.variable("x", shape1)
        y, lam_lb_y, lam_ub_y = nlp.variable("y", shape2, ub=10)
        _, lam_c1 = nlp.constraint("c1", x, op1, 5)
        _, lam_c2 = nlp.constraint("c2", 5, op2, y)

        dv = nlp.dual_variables
        c1_name = f"lam_{'g' if op1 == '==' else 'h'}_c1"
        c2_name = f"lam_{'g' if op2 == '==' else 'h'}_c2"
        self.assertTrue(cs.is_equal(dv["lam_lb_x"], lam_lb_x))
        self.assertTrue(cs.is_equal(dv["lam_ub_x"], lam_ub_x))
        self.assertTrue(cs.is_equal(dv["lam_lb_y"], lam_lb_y))
        self.assertTrue(cs.is_equal(dv["lam_ub_y"], lam_ub_y))
        self.assertTrue(cs.is_equal(dv[c1_name], lam_c1))
        self.assertTrue(cs.is_equal(dv[c2_name], lam_c2))

        if flag:
            lams = [lam_c1, lam_c2, lam_ub_y]
        else:
            lams = [lam_c1, lam_c2, lam_lb_x, lam_lb_y, lam_ub_x, lam_ub_y]
        actual_lam = cs.vertcat(*(cs.vec(o) for o in lams))
        expected_lam = nlp.lam
        self.cmp(actual_lam, expected_lam, vars=lams)

    def test_primal_dual_variables__returns_correctly(self):
        nlp = Nlp(sym_type=self.sym_type)
        x, lam_lbx, lam_ubx = nlp.variable("x", (2, 3), lb=[[0], [-np.inf]], ub=1)
        _, lam_g = nlp.constraint("c1", x[:, 0], "==", 2)
        _, lam_h = nlp.constraint("c2", x[0, :] + x[1, :] ** 2, "<=", 2)

        lam = cs.vertcat(lam_g, lam_h, lam_lbx, lam_ubx)
        y = cs.veccat(x, lam_g, lam_h, lam_lbx, lam_ubx)

        vars = [x, lam_g, lam_h, lam_lbx, lam_ubx]
        self.cmp(nlp.lam, lam, vars=vars)
        self.cmp(nlp.primal_dual, y, vars=vars)

    @parameterized.expand([(True,), (False,)])
    def test_h_lbx_ubx__returns_correct_indices(self, remove_red_bnds: bool):
        nlp = Nlp(sym_type=self.sym_type, remove_redundant_x_bounds=remove_red_bnds)

        x1, lam_lbx_c1, lam_ubx_c1 = nlp.variable("x1", (2, 1))
        lam_lbx1, lam_ubx1 = nlp.lam_lbx, nlp.lam_ubx
        h_lbx1, h_ubx1 = nlp.h_lbx, nlp.h_ubx
        vars1 = [x1, lam_lbx_c1, lam_ubx_c1]
        x2, lam_lbx_c2, lam_ubx_c2 = nlp.variable("x2", (2, 1), lb=0)
        lam_lbx2, lam_ubx2 = nlp.lam_lbx, nlp.lam_ubx
        h_lbx2, h_ubx2 = nlp.h_lbx, nlp.h_ubx
        vars2 = vars1.copy() + [x2, lam_lbx_c2, lam_ubx_c2]
        x3, lam_lbx_c3, lam_ubx_c3 = nlp.variable("x3", (2, 1), ub=1)
        lam_lbx3, lam_ubx3 = nlp.lam_lbx, nlp.lam_ubx
        h_lbx3, h_ubx3 = nlp.h_lbx, nlp.h_ubx
        vars3 = vars2.copy() + [x3, lam_lbx_c3, lam_ubx_c3]
        x4, lam_lbx_c4, lam_ubx_c4 = nlp.variable("x4", (2, 1), lb=0, ub=1)
        lam_lbx4, lam_ubx4 = nlp.lam_lbx, nlp.lam_ubx
        h_lbx4, h_ubx4 = nlp.h_lbx, nlp.h_ubx
        vars4 = vars3.copy() + [x4, lam_lbx_c4, lam_ubx_c4]

        if remove_red_bnds:
            self.assertTrue(
                all(o.is_empty() for o in [h_lbx1, lam_lbx1, h_ubx1, lam_ubx1])
            )
            self.assertTrue(all(o.is_empty() for o in [h_ubx2, lam_ubx2]))
            self.cmp(cs.evalf(h_lbx2 - (-x2)), 0, vars=vars2)
            self.cmp(cs.evalf(lam_lbx2 - lam_lbx_c2), 0, vars=vars2)
            self.cmp(cs.evalf(h_lbx3 - (-x2)), 0, vars=vars3)
            self.cmp(cs.evalf(lam_lbx3 - lam_lbx2), 0, vars=vars3)
            self.cmp((h_ubx3 - (x3 - 1)), 0, vars=vars3, force_numerical=True)
            self.cmp(cs.evalf(lam_ubx3 - lam_ubx_c3), 0, vars=vars3)
            self.cmp(h_lbx4 - cs.vertcat(-x2, -x4), 0, vars=vars4, force_numerical=True)
            e = lam_lbx4 - cs.vertcat(lam_lbx_c2, lam_lbx_c4)
            self.cmp(e, 0, vars=vars4, force_numerical=True)
            e = h_ubx4 - cs.vertcat(x3 - 1, x4 - 1)
            self.cmp(e, 0, vars=vars4, force_numerical=True)
            e = lam_ubx4 - cs.vertcat(lam_ubx_c3, lam_ubx_c4)
            self.cmp(e, 0, vars=vars4, force_numerical=True)
        else:
            self.cmp(cs.evalf(lam_lbx1 - lam_lbx_c1), 0, vars=vars1)
            self.cmp(cs.evalf(lam_ubx1 - lam_ubx_c1), 0, vars=vars1)
            e = lam_lbx2 - cs.vertcat(lam_lbx_c1, lam_lbx_c2)
            self.cmp(e, 0, vars=vars2, force_numerical=True)
            e = lam_ubx2 - cs.vertcat(lam_ubx_c1, lam_ubx_c2)
            self.cmp(e, 0, vars=vars2)
            e = lam_lbx3 - cs.vertcat(lam_lbx_c1, lam_lbx_c2, lam_lbx_c3)
            self.cmp(e, 0, vars=vars3, force_numerical=True)
            e = lam_ubx3 - cs.vertcat(lam_ubx_c1, lam_ubx_c2, lam_ubx_c3)
            self.cmp(e, 0, vars=vars3, force_numerical=True)
            e = lam_lbx4 - cs.vertcat(lam_lbx_c1, lam_lbx_c2, lam_lbx_c3, lam_lbx_c4)
            self.cmp(e, 0, vars=vars4, force_numerical=True)
            e = lam_ubx4 - cs.vertcat(lam_ubx_c1, lam_ubx_c2, lam_ubx_c3, lam_ubx_c4)
            self.cmp(e, 0, vars=vars4, force_numerical=True)

            if self.sym_type == "SX":
                self.cmp(h_lbx1 - (-np.inf - x1), 0, vars=vars1, force_numerical=True)
                self.cmp(h_ubx1 - (x1 - np.inf), 0, vars=vars1, force_numerical=True)
                e = h_lbx2 - cs.vertcat(-np.inf - x1, 0 - x2)
                self.cmp(cs.evalf(e), 0, vars=vars2)
                e = h_ubx2 - cs.vertcat(x1 - np.inf, x2 - np.inf)
                self.cmp(cs.evalf(e), 0, vars=vars2)
                e = h_lbx3 - cs.vertcat(-np.inf - x1, 0 - x2, -np.inf - x3)
                self.cmp(cs.evalf(e), 0, vars=vars3)
                e = h_ubx3 - cs.vertcat(x1 - np.inf, x2 - np.inf, x3 - 1)
                self.cmp(cs.evalf(e), 0, vars=vars3)
                e = h_lbx4 - cs.vertcat(-np.inf - x1, 0 - x2, -np.inf - x3, 0 - x4)
                self.cmp(cs.evalf(e), 0, vars=vars4)
                e = h_ubx4 - cs.vertcat(x1 - np.inf, x2 - np.inf, x3 - 1, x4 - 1)
                self.cmp(cs.evalf(e), 0, vars=vars4)

    def test_init_solver__raises__when_objective_not_set(self):
        nlp = Nlp(sym_type=self.sym_type)
        with self.assertRaises(RuntimeError):
            nlp.init_solver(OPTS)

    def test_init_solver__raises__when_type_is_wrong(self):
        nlp = Nlp(sym_type=self.sym_type)
        with self.assertRaises(ValueError):
            nlp.init_solver(OPTS, type="a_random_type")

    def test_init_solver__saves_options_correctly(self):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        nlp.minimize(x**2)
        nlp.init_solver(OPTS)
        self.assertDictEqual(OPTS, nlp.solver_opts)

    @parameterized.expand([("sqpmethod",), ("osqp",)])
    def test_init_solver__chooses_conic_or_nlp_correctly(self, solver: str):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        nlp.minimize(x**2)

        old_qpsol, old_nlpsol = cs.qpsol, cs.nlpsol
        cs.qpsol = mock_qpsol = Mock()
        cs.nlpsol = mock_nlpsol = Mock()
        try:
            nlp.init_solver(solver=solver)
        finally:
            cs.qpsol, cs.nlpsol = old_qpsol, old_nlpsol

        if solver == "sqpmethod":
            mock_qpsol.assert_not_called()
            mock_nlpsol.assert_called_once()
        else:
            mock_qpsol.assert_called_once()
            mock_nlpsol.assert_not_called()

    def test_solve__raises__with_uninit_solver(self):
        nlp = Nlp(sym_type=self.sym_type)
        with self.assertRaisesRegex(RuntimeError, "Solver uninitialized."):
            nlp.solve(None)

    def test_solve__raises__with_free_parameters(self):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        p = nlp.parameter("p")
        nlp.minimize(p * (x**2))
        nlp.init_solver(OPTS)
        with self.assertRaisesRegex(
            RuntimeError, "Trying to solve the NLP with unspecified parameters: p."
        ):
            nlp.solve({})

    def test_solve__computes_correctly__example_0(self):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", (2, 1))[0]
        y = nlp.variable("y", (3, 1))[0]
        p = nlp.parameter("p")
        nlp.minimize(p + (x.T @ x + y.T @ y))
        nlp.init_solver(OPTS)
        sol = nlp.solve({"p": 3})
        self.assertTrue(sol.success)
        np.testing.assert_allclose(sol.f, 3)
        for v in sol.vals.values():
            np.testing.assert_allclose(v, 0)
        o = sol.value(p + (x.T @ x + y.T @ y))
        np.testing.assert_allclose(sol.f, o)

    def test_solve__computes_corretly__example_1a(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_1
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        y = nlp.variable("y")[0]
        nlp.constraint("c1", x**2 + y**2, "==", 1)
        nlp.minimize(-x - y)
        nlp.init_solver(OPTS)
        sol = nlp.solve()
        np.testing.assert_allclose(-sol.f, np.sqrt(2))
        for k in ("x", "y"):
            np.testing.assert_allclose(sol.vals[k], np.sqrt(2) / 2, atol=1e-9)
        np.testing.assert_allclose(sol.value(nlp.lam_g), 1 / np.sqrt(2), atol=1e-9)

    def test_solve__computes_corretly__example_1b(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_1
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        y = nlp.variable("y")[0]
        nlp.constraint("c1", x**2 + y**2, "==", 1)
        nlp.minimize((x + y) ** 2)
        nlp.init_solver(OPTS)
        sol = nlp.solve(vals0={"x": 0.5, "y": np.sqrt(1 - 0.5**2)})
        np.testing.assert_allclose(sol.f, 0, atol=1e-9)
        np.testing.assert_allclose(abs(sol.vals["x"]), np.sqrt(2) / 2, atol=1e-9)
        np.testing.assert_allclose(abs(sol.vals["y"]), np.sqrt(2) / 2, atol=1e-9)
        np.testing.assert_allclose(sol.value(nlp.lam_g), 0, atol=1e-9)

    def test_solve__computes_corretly__example_2(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_2
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        y = nlp.variable("y")[0]
        nlp.constraint("c1", x**2 + y**2, "==", 3)
        nlp.minimize(x**2 * y)
        nlp.init_solver(OPTS)
        sol = nlp.solve(vals0={"x": np.sqrt(3 - 0.8**2), "y": -0.8})
        np.testing.assert_allclose(sol.f, -2, atol=1e-9)
        np.testing.assert_allclose(abs(sol.vals["x"]), np.sqrt(2), atol=1e-9)
        np.testing.assert_allclose(sol.vals["y"], -1, atol=1e-9)
        np.testing.assert_allclose(
            sol.value(sol.vals["x"] * (sol.vals["y"] + nlp.lam_g)), 0, atol=1e-9
        )

    def test_solve__computes_corretly__example_3(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_3
        n = 50
        nlp = Nlp(sym_type=self.sym_type)
        p = nlp.variable("p", (n, 1), lb=1 / n * 1e-6)[0]
        nlp.constraint("c1", cs.sum1(p), "==", 1)
        nlp.minimize(cs.sum1(p * log(p, 2)))
        nlp.init_solver(OPTS)
        sol = nlp.solve(vals0={"p": np.random.rand(n)})
        np.testing.assert_allclose(sol.vals["p"], 1 / n, atol=1e-9)
        np.testing.assert_allclose(
            sol.value(-(1 / cs.log(2) + log(p, 2)) - nlp.lam_g), 0, atol=1e-6
        )

    def test_solve__computes_corretly__example_4(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_4
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x")[0]
        nlp.constraint("c1", x**2, "==", 1)
        nlp.minimize(x**2)
        nlp.init_solver(OPTS)
        sol = nlp.solve(vals0={"x": 1 + 0.1 * np.random.rand()})
        np.testing.assert_allclose(sol.f, 1, atol=1e-9)
        np.testing.assert_allclose(sol.value(nlp.lam_g), -1, atol=1e-9)

    def test_solve__computes_corretly__example_5(self):
        # https://personal.math.ubc.ca/~israel/m340/kkt2.pdf
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", (2, 1), lb=0)[0]
        nlp.constraint("c1", x[0] + x[1] ** 2, "<=", 2)
        nlp.minimize(-x[0] * x[1])
        nlp.init_solver(OPTS)
        sol = nlp.solve(vals0={"x": [1, 1]})
        np.testing.assert_allclose(-sol.f, np.sqrt(2 / 3) * 4 / 3, atol=1e-9)
        np.testing.assert_allclose(
            sol.vals["x"].full().flatten(), [4 / 3, np.sqrt(2 / 3)], atol=1e-9
        )
        np.testing.assert_allclose(
            sol.value(cs.vertcat(nlp.lam_h, nlp.lam_lbx)).full().flatten(),
            [np.sqrt(2 / 3), 0, 0],
            atol=1e-7,
        )

    def test_solve__computes_corretly__example_6(self):
        # https://www.reddit.com/r/cheatatmathhomework/comments/sw6nqs/optimization_using_kkt_conditions_for_constraints/
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", (3, 1), lb=0)[0]
        y = nlp.variable("y", (3, 1), lb=0)[0]
        z = nlp.variable("z", (3, 1), lb=0)[0]
        nlp.minimize(cs.sum1(cs.vertcat(x, y, z)))
        nlp.constraint("c1", x.T @ [25, 15, 75] + 3 * cs.sqrt(cs.sumsqr(x)), "<=", 56)
        nlp.constraint("c2", y.T @ [75, 3, 3] + 3 * cs.sqrt(cs.sumsqr(y)), "<=", 87)
        nlp.constraint("c3", z.T @ [15, 22, 4] + 3 * cs.sqrt(cs.sumsqr(z)), "<=", 38)
        nlp.init_solver(OPTS)
        sol = nlp.solve()
        for name, v in [("x", x), ("y", y), ("z", z)]:
            val1 = sol.vals[name]
            val2 = sol.value(v)
            val3 = sol.value(nlp.variables[name])
            val4 = sol.value(nlp._vars[name])
            np.testing.assert_allclose(val1, val2, atol=1e-9)
            np.testing.assert_allclose(val2, val3, atol=1e-9)
            np.testing.assert_allclose(val3, val4, atol=1e-9)

    def test_to_function__raises__with_uninitiliazed_solver(self):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", lb=0)[0]
        y = nlp.variable("y")[0]
        xy = cs.vertcat(x, y)
        p = nlp.parameter("p")
        with self.assertRaises(RuntimeError):
            nlp.to_function("M", [p, xy], [xy], ["p", "xy"], ["xy"])

    def test_to_function__raises__with_free_variables(self):
        a = 0.2
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", lb=0)[0]
        y = nlp.variable("y")[0]
        xy = cs.vertcat(x, y)
        c = nlp._sym_type.sym("c")
        p = nlp.parameter("p")
        nlp.minimize((1 - x) ** 2 + a * (y - x**2) ** 2)
        nlp.init_solver(OPTS)
        with self.assertRaises(RuntimeError):
            nlp.to_function("M", [p], [xy], ["xy"], ["xy"])
        with self.assertRaises(RuntimeError):
            nlp.to_function("M", [p, xy], [xy, c], ["xy"], ["xy"])

    @parameterized.expand([(False,), (True,)])
    def test_to_function__computes_correct_solution__also_with_deepcopy(
        self, copy: bool
    ):
        a = 0.2
        nlp = Nlp(sym_type=self.sym_type)
        if copy:
            nlp = deepcopy(nlp)
        x = nlp.variable("x", lb=0)[0]
        y = nlp.variable("y")[0]
        xy = cs.vertcat(x, y)
        p = nlp.parameter("p")
        nlp.minimize((1 - x) ** 2 + a * (y - x**2) ** 2)
        g = (x + 0.5) ** 2 + y**2
        nlp.constraint("c1", (p / 2) ** 2, "<=", g)
        nlp.constraint("c2", g, "<=", p**2)
        nlp.init_solver(OPTS)

        M = nlp.to_function("M", [p, xy], [xy], ["p", "xy"], ["xy_opt"])

        sol = nlp.solve(pars={"p": 1.25})
        xy1 = sol.value(xy).full().flatten()
        xy2 = M(1.25, 0).full().flatten()

        np.testing.assert_allclose(xy1, [0.719011, 0.276609], atol=1e-4)
        np.testing.assert_allclose(xy2, [0.719011, 0.276609], atol=1e-4)

    def test_can_be_pickled(self):
        a = 0.2
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", lb=0)[0]
        y = nlp.variable("y")[0]
        p = nlp.parameter("p")
        nlp.minimize((1 - x) ** 2 + a * (y - x**2) ** 2)
        g = (x + 0.5) ** 2 + y**2
        nlp.constraint("c1", (p / 2) ** 2, "<=", g)
        nlp.constraint("c2", g, "<=", p**2)
        nlp.init_solver(OPTS)

        with cs.global_pickle_context():
            pickled = pickle.dumps(nlp)
        with cs.global_unpickle_context():
            other = pickle.loads(pickled)

        self.assertEqual(nlp.name, other.name)

    @parameterized.expand(product(("both", "lb", "ub"), (True, False)))
    def test_remove_variable_bounds__remove_bounds_correctly(
        self, direction: str, all_idx: bool
    ):
        shape = tuple(np.random.randint(3, 10, size=2))
        lb = np.random.rand(*shape) - 3
        ub = np.random.rand(*shape) + 3
        if all_idx:
            idx_to_remove = list(product(range(shape[0]), range(shape[1])))
        else:
            n_to_remove = np.random.randint(1, np.prod(shape) // 2)
            idx_to_remove = np.random.randint((0, 0), shape, size=(n_to_remove, 2))

        nlp = Nlp(sym_type=self.sym_type, remove_redundant_x_bounds=True)
        u_size = np.prod(nlp.variable("u", (5, 2), ub=+1)[0].shape)  # to create noise
        nlp.variable("x", shape, lb=lb, ub=ub)
        nlp.variable("z", (7, 9), lb=+2, ub=+3)  # to create noise
        nlp.remove_variable_bounds("x", direction, None if all_idx else idx_to_remove)

        if direction in {"both", "lb"}:
            lb_ = lb.copy()
            lb_mask_ = np.full(lb.shape, False)
            for i in idx_to_remove:
                lb_[tuple(i)] = -np.inf
                lb_mask_[tuple(i)] = True
            exp_lb = nlp.lbx.data[u_size : u_size + np.prod(shape)]
            exp_lb_mask = nlp.lbx.mask[u_size : u_size + np.prod(shape)]
            np.testing.assert_array_equal(lb_.reshape(-1, order="F"), exp_lb)
            np.testing.assert_array_equal(lb_mask_.reshape(-1, order="F"), exp_lb_mask)
            self.assertTrue(
                nlp.dual_variables["lam_lb_x"].size1() == (~exp_lb_mask).sum()
            )
            self.assertTrue(nlp.h_lbx.shape == nlp.lam_lbx.shape)
        if direction in {"both", "ub"}:
            ub_ = ub.copy()
            ub_mask_ = np.full(ub.shape, False)
            for i in idx_to_remove:
                ub_[tuple(i)] = +np.inf
                ub_mask_[tuple(i)] = True
            exp_ub = nlp.ubx.data[u_size : u_size + np.prod(shape)]
            exp_ub_mask = nlp.ubx.mask[u_size : u_size + np.prod(shape)]
            np.testing.assert_array_equal(ub_.reshape(-1, order="F"), exp_ub)
            np.testing.assert_array_equal(ub_mask_.reshape(-1, order="F"), exp_ub_mask)
            self.assertTrue(
                nlp.dual_variables["lam_ub_x"].size1() == (~exp_ub_mask).sum()
            )
            self.assertTrue(nlp.h_ubx.shape == nlp.lam_ubx.shape)

    @parameterized.expand(product(("both", "g", "h"), (False, True)))
    def test_remove_constraints__remove_bounds_correctly(
        self, remove: str, all_idx: bool
    ):
        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", tuple(np.random.randint(2, 5, size=2)))[0]
        y = nlp.variable("y", tuple(np.random.randint(2, 5, size=2)))[0]
        z = nlp.variable("z", tuple(np.random.randint(2, 5, size=2)))[0]  # noise
        w = nlp.variable("w", tuple(np.random.randint(2, 5, size=2)))[0]  # noise
        q = nlp.variable("q", tuple(np.random.randint(2, 5, size=2)))[0]  # noise
        p = nlp.variable("p", tuple(np.random.randint(2, 5, size=2)))[0]  # noise
        nlp.minimize(cs.sumsqr(cs.veccat(x, y, z, w, q, p)))
        nlp.constraint("h0", z, ">=", 0.0)  # noise
        nlp.constraint("h1", x, ">=", 1.0)
        nlp.constraint("h2", q, ">=", 0.0)  # noise
        nlp.constraint("g0", w, "==", 0.0)  # noise
        nlp.constraint("g1", y, "==", 2.0)
        nlp.constraint("g2", p, "==", 0.0)  # noise
        nlp.init_solver(OPTS)

        if all_idx:
            idx_to_remove_h1 = idx_to_remove_g1 = None
        else:
            n_to_remove = np.random.randint(1, np.prod(x.shape) // 2)
            idx_to_remove_h1 = np.unique(
                np.random.randint((0, 0), x.shape, size=(n_to_remove, 2)), axis=0
            )
            n_to_remove = np.random.randint(1, np.prod(y.shape) // 2)
            idx_to_remove_g1 = np.unique(
                np.random.randint((0, 0), y.shape, size=(n_to_remove, 2)), axis=0
            )
        remove_h = remove in {"both", "h"}
        remove_g = remove in {"both", "g"}
        if remove_h:
            nlp.remove_constraints("h1", idx_to_remove_h1)
        if remove_g:
            nlp.remove_constraints("g1", idx_to_remove_g1)

        sol = nlp.solve()

        expected_nh = cs.veccat(z, x, q).size1()
        expected_ng = cs.veccat(w, y, p).size1()
        expected_x = np.full(x.shape, 1.0)
        expected_y = np.full(y.shape, 2.0)
        if remove_h:
            if all_idx:
                expected_nh -= np.prod(x.shape)
                expected_x.fill(0.0)
            else:
                expected_nh -= idx_to_remove_h1.shape[0]
                for idx in idx_to_remove_h1:
                    expected_x[idx[0], idx[1]] = 0.0
        if remove_g:
            if all_idx:
                expected_ng -= np.prod(y.shape)
                expected_y.fill(0.0)
            else:
                expected_ng -= idx_to_remove_g1.shape[0]
                for idx in idx_to_remove_g1:
                    expected_y[idx[0], idx[1]] = 0.0

        self.assertEqual(nlp.nh, expected_nh)
        self.assertEqual(nlp.ng, expected_ng)
        np.testing.assert_allclose(sol.vals["z"], 0.0, atol=1e-4, rtol=1e-4)
        np.testing.assert_allclose(sol.vals["x"], expected_x, atol=1e-4, rtol=1e-4)
        np.testing.assert_allclose(sol.vals["q"], 0.0, atol=1e-4, rtol=1e-4)
        np.testing.assert_allclose(sol.vals["w"], 0.0, atol=1e-4, rtol=1e-4)
        np.testing.assert_allclose(sol.vals["y"], expected_y, atol=1e-4, rtol=1e-4)
        np.testing.assert_allclose(sol.vals["p"], 0.0, atol=1e-4, rtol=1e-4)


if __name__ == "__main__":
    unittest.main()



r"""A module with utility functions and classes around and for optimization.

Overview
========

It contains the following submodules:

- :mod:`csnlp.util.docs`: a collection of stand-alone functions to extract information
  from the CasADi documentation via code. In particular, this module offers a way to get
  the solvers that are available in CasADi (i.e., they have an interface) as well as and
  their options. The functions are taken from the
  `MPCTools <https://bitbucket.org/rawlings-group/mpc-tools-casadi/src/master/mpctools/util.py>`_
  repository by the Rawlings' group.
- :mod:`csnlp.util.io`: a collection of utilities for input/output operations. The goals
  of these methods are:

   * compatibility of pickling/deepcopying with CasADi objects and classes that hold
     such objects (since these are often not picklable)
   * saving and loading data to/from files, possibly compressed.

- :mod:`csnlp.util.math`: a collection of stand-alone functions that implement some of
  the basic mathematical operations that are not available in CasADi. The
  implementations are simple and thus not optimized for performance. They are meant to
  be used as a fallback when the CasADi ally does not provide the required
  functionality.

Submodules
==========

.. autosummary::
   :toctree: generated
   :template: module.rst

   docs
   io
   math
"""

__all__ = ["docs", "io", "math"]

from . import docs, io, math



"""A collection of two methods for computing higher-order sensitivities (i.e., Jacobian
and Hessian) w.r.t. CasADi symbolic variables. Natively, CasADi does not support
jacobian or hessian for matrices (or at least, they will be flattened). These
"higher-order" functions allows to compute the jacobian and hessian of a matrix w.r.t.
another matrix."""

from itertools import product as _product
from typing import Union

import casadi as cs
import numpy as np

from .data import array2cs as _array2cs
from .data import cs2array as _cs2array


def hojacobian(ex: Union[cs.MX, cs.SX], x: Union[cs.MX, cs.SX]) -> np.ndarray:
    """Computes jacobian on higher-order matrices, not just vectors.

    Parameters
    ----------
    ex : casadi.MX or SX
        The expression to compute the jacobian for. Can be a matrix.
    x : casadi.MX or SX
        The variable to differentiate with respect to. Can be a matrix.

    Returns
    -------
    numpy array of symbolic variables
        A 4D array of objects, where each entry ``(i,j,k,m)`` is the derivative
        of ``ex[i,j]`` w.r.t. ``x[k,m]``.
    """
    return _cs2array(cs.jacobian(cs.vec(ex), cs.vec(x))).reshape(
        ex.shape + x.shape, order="F"
    )


def hohessian(
    ex: Union[cs.MX, cs.SX], x: Union[cs.MX, cs.SX], y: Union[cs.MX, cs.SX, None] = None
) -> tuple[np.ndarray, np.ndarray]:
    """Computes hessian on higher-order matrices, similar to :func:`hojacobian`.

    Parameters
    ----------
    ex : casadi.MX or SX
        The expression to compute the hessin for. Can be a matrix.
    x : casadi.MX or SX
        The variable to differentiate with respect to. Can be a matrix.
    y : casadi.MX or SX, optional
        Use this argument to specify the second partial derivative, i.e., if
        the output requested is not the hessian of ``ex`` w.r.t. ``x``, but rather
        the second-order partial derivatives of ``ex`` w.r.t. ``x`` and ``y``.

    Returns
    -------
    tuple of 2 numpy arrays of symbolic variables
        The first element is a 6D array of objects, where each entry
        ``(i,j,k,m,l,p)`` is the hessian of ``ex[i,j]`` w.r.t. ``x[k,m], y[l,p]``,
        while the second element contains the jacobian instead (see :func:`hojacobian`
        for details).
    """
    if y is None:
        y = x
    J = hojacobian(ex, x)
    H = np.empty(ex.shape + x.shape + y.shape, object)
    for i in _product(*map(range, ex.shape)):
        H[i] = hojacobian(_array2cs(J[i]), y)
    return H, J



from math import prod
from typing import Literal, TypeVar

import casadi as cs
import numpy as np
from numpy.typing import NDArray

from .parameters import HasParameters

SymType = TypeVar("SymType", cs.SX, cs.MX)


class HasVariables(HasParameters[SymType]):
    """Class for the creation and storage of symbolic variables in an NLP problem. It
    builds on top of :class:`HasParameters`, which handles parameters.

    Parameters
    ----------
    sym_type : {"SX", "MX"}, optional
        The CasADi symbolic variable type to use in the NLP, by default ``"SX"``.
    """

    def __init__(self, sym_type: Literal["SX", "MX"] = "SX") -> None:
        super().__init__(sym_type)
        self._vars: dict[str, SymType] = {}
        self._x = self._sym_type()
        self._discrete = np.empty(0, dtype=bool)

    @property
    def x(self) -> SymType:
        """Gets the primary variables of the NLP scheme in vector form."""
        return self._x

    @property
    def nx(self) -> int:
        """Number of variables in the NLP scheme."""
        return self._x.shape[0]

    @property
    def variables(self) -> dict[str, SymType]:
        """Gets the primal variables of the NLP scheme."""
        return self._vars

    @property
    def discrete(self) -> NDArray[np.bool_]:
        """Gets the boolean array indicating which variables are discrete."""
        return self._discrete

    def variable(
        self,
        name: str,
        shape: tuple[int, int] = (1, 1),
        discrete: bool = False,
    ) -> SymType:
        """Adds a variable to the NLP problem.

        Parameters
        ----------
        name : str
            Name of the new variable. Must not be already in use.
        shape : tuple[int, int], optional
            Shape of the new parameter. By default a scalar, i.e., ``(1, 1)``.
        discrete : bool, optional
            Flag indicating if the variable is discrete. Defaults to ``False``.

        Returns
        -------
        var : casadi.SX
            The symbol of the new variable.

        Raises
        ------
        ValueError
            Raises if there is already another variable with the same name ``name``.
        """
        if name in self._vars:
            raise ValueError(f"Variable name '{name}' already exists.")
        var = self._sym_type.sym(name, *shape)
        self._vars[name] = var
        self._x = cs.veccat(self._x, var)
        self._discrete = np.concatenate(
            (self._discrete, np.full(prod(shape), discrete, dtype=bool))
        )
        return var
